/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "fcintl.h"
#include "game.h"
#include "log.h"
#include "map.h"
#include "mem.h"
#include "rand.h"
#include "shared.h"
#include "support.h"

#include "mapgen.h"

/*  --- cut here and move to map.h ---  */

/* Replaces these definitions in map.h */
#undef  cartesian_adjacent_iterate
#undef  cartesian_adjacent_iterate_end
#undef  adjc_iterate
#undef  adjc_iterate_end

/* Macros for moving around the 1st and 2nd square in NWSE directions
 * There are 8n directions or edge tiles in the n'th square outwards,
 *   (2n+1)^2 => 1, 9, 25 sized squares, (2n+1)^2 - (2n-1)^2= 8n.
 *
 * The Delta macros step to direction "n", wrapping at high/low ends.
 *
 * Iterate macros below use a single indexed while loop based on the
 *   following known ordering, with access to the index (i.e. _nn).
 *   Example: if _nn= 0, you are NW of your start.
 *
 * All allocate the user's chosen x, y variables within the (core) loop.
 * All treat arguments other than x, y as single use expressions.
 * All use a single loop on the index value _nn.
 * All blocks will be correctly wrapped in the x dimension at the
 *  map.xsize boundary, and off map y values will be skipped.
 *
 * Depends on map.xsize, map.ysize, map_adjust_x()
*/

/* Array of offsets starting at the NW corner and moving clockwise,   */
/* including the origin. Note, first 1, 9, and 25 == square iteration */
static int dx[] = { 0,
   -1,  0,  1,  1,  1,  0, -1, -1,
   -2, -1,  0,  1,  2,  2,  2,  2,  2,  1,  0, -1, -2, -2, -2, -2
};
static int dy[] = { 0,
   -1, -1, -1,  0,  1,  1,  1,  0,
   -2, -2, -2, -2, -2, -1,  0,  1,  2,  2,  2,  2,  2,  1,  0, -1
};

/* Delta macros - step to direction "n" in the 1st or 2nd square */
#define DX1(n)  dx[((n)&07)+1]
#define DY1(n)  dy[((n)&07)+1]
#define DX2(n)  dx[((n)&15)+9]
#define DY2(n)  dy[((n)&15)+9]


/* Core adjacent iteration loop 1st square (+1) */
#define core_iterate(ni, n0, x0, y0, x, y) {            \
  int x, y, _ni= (ni), _nn= (n0), _x0= (x0), _y0= (y0); \
  while( (_nn-= _ni) >= 0 ) {                           \
    y= dy[_nn+1]+_y0;                                   \
    if( y < 0 || y >= map.ysize ) continue;             \
    x= dx[_nn+1]+_x0, x= map_adjust_x(x);

#define core_iterate_end                                \
  }                                                     \
}

/* Core adjacent iteration loop 2nd square (+9) */
#define core_iterate2(ni, n0, x0, y0, x, y) {           \
  int x, y, _ni= (ni), _nn= (n0), _x0= (x0), _y0= (y0); \
  while( (_nn-= _ni) >= 0 ) {                           \
    y= dy[_nn+9]+_y0;                                   \
    if( y < 0 || y >= map.ysize ) continue;             \
    x= dx[_nn+9]+_x0, x= map_adjust_x(x);

#define core_iterate2_end                               \
  }                                                     \
}

/* Cartesian W, S, E, N iteration */
#define cartesian_adjacent_iterate(x0, y0, x, y) {      \
  core_iterate(2, 9, x0, y0, x, y)

#define cartesian_adjacent_iterate_end                  \
  core_iterate_end                                      \
}

/* Cartesian diagonal iteration SW, SE, NE, NW */
#define cartesian_diagonal_iterate(x0, y0, x, y) {      \
  core_iterate(2, 8, x0, y0, x, y)

#define cartesian_diagonal_iterate_end                  \
  core_iterate_end                                      \
}

/* full 1st square iteration W, SW, ... NW */
#define adjc_iterate(x0, y0, x, y) {                    \
  core_iterate(1, 8, x0, y0, x, y)

#define adjc_iterate_end                                \
  core_iterate_end                                      \
}


/* Cartesian W, S, E, N iteration: 2nd square*/
#define cartesian_adjacent_iterate2(x0, y0, x, y) {     \
  core_iterate2(4, 18, x0, y0, x, y)

#define cartesian_adjacent_iterate2_end                 \
  core_iterate2_end                                     \
}

/* Cartesian diagonal iteration: 2nd square */
#define cartesian_diagonal_iterate2(x0, y0, x, y) {     \
  core_iterate2(4, 16, x0, y0, x, y)

#define cartesian_diagonal_iterate2_end                 \
  core_iterate2_end                                     \
}

/* Cartesian knights iteration: 2nd square */
#define cartesian_knights_iterate2(x0, y0, x, y) {      \
  core_iterate2(2, 17, x0, y0, x, y)

#define cartesian_knights_iterate2_end                  \
  core_iterate2_end                                     \
}

/* full iteration: 2nd square */
#define adjc_iterate2(x0, y0, x, y) {                   \
  core_iterate2(1, 16, x0, y0, x, y)

#define adjc_iterate2_end                               \
  core_iterate2_end                                     \
}
/* 2001/06/14 -rww */

/*  --- end cut for move to map.h ---  */


/*  --- cut here and move to map.h ---  */

/* Replaces these definitions in map.h */
#undef  square_iterate
#undef  square_iterate_end
#undef  whole_map_iterate
#undef  whole_map_iterate_end

/*
 * block_iterate(x,y,x0,y0,dx,dy) loops the user variables x and y
 *   over the tiles in the rectangular block with corner x0, y0 and
 *   dimensions dx, dy. Reasonable argument checking is done.
 *
 * block_iterate2(x,y,x0,y0,x1,y1) loops the user variables x and y
 *   over the tiles in the rectangular block defined by the two points
 *   x0, y0 and x1, y1. Reasonable argument checking is done.
 *
 * fastblock_iterate(x,y,xy1,xy0,xs,xm) is the unchecked core loop
 *   from point xy1 to xy0 with skips of xs whenever x equals xm.
 *
 * *** Warning: If you cannot figure out what block_iterate does,
 *              do not use fastblock_iterate directly.
 *
 * All allocate the user's chosen x, y variables within the (core) loop.
 * All treat arguments other than x, y as single use expressions.
 * All use a single loop on the linear offset _xy = map-xsize*y + x.
 * All blocks will be correctly wrapped in the x dimension at the
 *  map.xsize boundary, and off map y values will be skipped.
 *
 * Depends on map.xsize, map.ysize, map_adjust_x()
 */
#define fastblock_iterate(x, y, xy1, xy0, xs, xm) {                     \
  int x, y, _xm= (xm), _xs= (xs), _xy0= (xy0), _xy= (xy1);              \
  /* loop from upper xy1 to lower xy0 point skipping xs at xm */        \
  while( _xy-- > _xy0 ) {                                               \
    y= _xy/map.xsize, x= _xy%map.xsize;                                 \
    if( x == _xm ) { _xy-= _xs; continue; }

#define fastblock_iterate_end                                           \
  }                                                                     \
}


#define block_iterate(x, y, x0, y0, dx, dy) {                           \
  int _x0= (x0), _dx= (dx), _y0= (y0), _dy= (dy);                       \
  /* fixup y dimensions                                              */ \
  if( _dy < 0 ) _y0+= _dy+1, _dy= -_dy;          /* negative delta y */ \
  if((_dy+= _y0-1) >= map.ysize) _dy= map.ysize-1;  /* y1>=map.ysize */ \
  if( _y0 < 0 ) _y0= 0;                                    /* y0 < 0 */ \
  /* fixup x dimensions                                              */ \
  if( _dx < 0 ) _x0+= _dx+1, _dx= -_dx;          /* negative delta x */ \
  _x0= map_adjust_x(_x0);                                /* adjust x */ \
  /* set start and skip points taking wrap into account              */ \
  _dy*= map.xsize, _y0*= map.xsize;                                     \
  if( _dx >= map.xsize )                                                \
    _dy+= map.xsize, _dx= 0, _x0= map.xsize+1;            /* no skip */ \
  else                                                                  \
  if( _x0+_dx > map.xsize )                                             \
    _dy+= map.xsize, _dx=(map.xsize-1-_dx);       /* skip in midline */ \
  else                                          /* skip low and wrap */ \
    _y0+= _x0, _dy+= _x0+_dx-1, _dx=(map.xsize-1-_dx);                  \
  _x0= (_x0 == 0 ? map.xsize:_x0)-1;     /* skip on 1st out-of-bound */ \
  /* Note: _args carefully do not overlap with fastblock_iterate     */ \
  fastblock_iterate(x, y, _dy, _y0, _dx, _x0)

#define block_iterate_end                                               \
  fastblock_iterate_end                                                 \
}


#define block_iterate2(x, y, x0, y0, x1, y1) {                          \
  int _x0= (x0), _x1= (x1), _y0= (y0), _y1= (y1);                       \
  /* fixup y dimensions                                              */ \
  if( _y1 < _y0 ) _y0^= _y1, _y1^= _y0, _y0^= _y1;       /* flip y's */ \
  if( _y1 >= map.ysize) _y1= map.ysize-1;           /* y1>=map.ysize */ \
  if( _y0 < 0 ) _y0= 0;                                    /* y0 < 0 */ \
  /* fixup x dimensions                                              */ \
  _x1-= _x0, _x1= map_adjust_x(_x1)+1;                /* adjusted dx */ \
  _x0= map_adjust_x(_x0);                               /* adjust x0 */ \
  /* set start and skip points taking wrap into account              */ \
  _y1*= map.xsize, _y0*= map.xsize;                                     \
  if( _x1 >= map.xsize )                                                \
    _y1+= map.xsize, _x1= 0, _x0= map.xsize+1;            /* no skip */ \
  else                                                                  \
  if( _x0+_x1 > map.xsize )                                             \
    _y1+= map.xsize, _x1=(map.xsize-1-_x1);       /* skip in midline */ \
  else                                          /* skip low and wrap */ \
    _y0+= _x0, _y1+= _x0+_x1-1, _x1=(map.xsize-1-_x1);                  \
  _x0= (_x0 == 0 ? map.xsize:_x0)-1;     /* skip on 1st out-of-bound */ \
  /* Note: _args carefully do not overlap with fastblock_iterate     */ \
  fastblock_iterate(x, y, _y1, _y0, _x1, _x0)

#define block_iterate2_end                                              \
  fastblock_iterate_end                                                 \
}

/* backwards compatible square_iterate
 * It too uses a single loop on the linear offset _xy = map-xsize*y + x.
 */
#define square_iterate(x, y, n, x1, y1) {                               \
  int _nn= (n);                                                         \
  /* Note: _args carefully do not overlap with block_iterate */         \
  block_iterate(x1, y1, (x)-_nn, (y)-_nn, _nn+_nn+1, _nn+_nn+1)

#define square_iterate_end                                              \
  block_iterate_end                                                     \
}


/* optimized whole map iterate (fixed endpoints, no skip logic)
 * It too uses a single loop on the linear offset _xy = map-xsize*y + x.
 */
#define whole_map_iterate(x,y) {                                        \
  int x, y, _xy= map.ysize*map.xsize;                                   \
  while( _xy-- > 0 ) {                                                  \
    y= _xy/map.xsize, x= _xy%map.xsize;

#define whole_map_iterate_end                                           \
  }                                                                     \
}
/* 2001/06/14 -rww */

/*  --- end cut for move to map.h ---  */


/* Wrapper for ocean map - these should be in map.c with the corresponding
 * fields added to the map struct to parallel continent handling, plus
 * map.num_oceans.
 * Note: make sure ocean_map is allocated before calling these
*/
#define map_get_ocean(x,y)     ocean_map[(y)*map.xsize+map_adjust_x(x)]
#define map_set_ocean(x,y,t)  (ocean_map[(y)*map.xsize+map_adjust_x(x)]= (t))
#define map_num_oceans()  num_oceans

static int num_oceans, *ocean_map;

/* Wrapper for easy height_map access. It's a macro so it can be an lvalue. */
#define hmap(x,y) height_map[(y)*map.xsize+map_adjust_x(x)]

static int *height_map;


/* utility routines */
static int  hmap1(int x, int y, int n);
static int  hslope(int *n, int x, int y, int hi);
static int  is_map1(int x, int y, int n, int tt);
static int  is_rough(int x, int y);
static int  is_cold(int x, int y);
static int  map_dist(int p0, int p1);

/* core map generators */
static void mapgenerator1(void);
static void mapgenerator2(void);
static void mapgenerator3(void);
static void mapgenerator4(void);

static int  assign_continent_flood(int x, int y, int nr);
static int  assign_continent_flood1(int x, int y, int n0, int nr);
static int  list_continent_flood1(int *list, int x, int y, int n0);
static void setup_isledata(void);
static int  assign_ocean_flood(int x, int y, int nr);
static int  assign_ocean_flood1(int x, int y, int n0, int nr);
static int  list_ocean_flood(int *list, int x, int y);
static int  list_ocean_flood1(int *list, int x, int y, int n0);
static void setup_seadata(void);
static void map_check(void);
static void add_specials(int prob);
static void make_huts(int number);

/* mapgenerator1 routines */
static void adjust_map2(void);
static void adjust_map(void);
static void smooth_map(void);

static void make_mountains(int hill);
static void make_polar(void);
static int  make_desert(int x, int y, int height, int diff) ;
static void make_deserts(void);
static void make_forest(int x, int y, int height, int diff);
static void make_forests(void);
static void make_swamps(void);
static int  make_river(int x, int y, int force) ;
static void make_rivers(void);
static void make_plains(void);
static void make_passable(void);
static void make_fair(void);
static void make_land(void);
static int  is_tiny_island(int x, int y) ;
static void remove_tiny_islands(void);
static int  is_tiny_lake(int x, int y) ;
static void remove_tiny_lakes(void);

/* mapgenerator2-4 routines */
static void initworld(void);
static void make_island(int islemass, int starters);
static void create_island(int *height_map, int islemass);
static int  place_island(int *create_map, int forever);
static void fill_island(long int *bucket, int coast,
	int warm0_weight, int warm1_weight, int cold0_weight, int cold1_weight,
	enum tile_terrain_type warm0, enum tile_terrain_type warm1,
	enum tile_terrain_type cold0, enum tile_terrain_type cold1);

/* debug */
static void print_hmap(int *height_map);
static void print_tmap(void);
static void print_fmap(void);

/* global data */
static int maxval=10000;
static int thill=6000;
static int tocean=3000;
static int trough=600;
static int forests=0;

struct isledata {
  int x,y;			  /* upper left corner of the islands */
  int goodies;
  int starters;
};
static struct isledata *islands;

struct seadata {
  int x, y;			 /* upper left corner of the islands */
  int count;			/* number of connected water tiles  */
  int rivers;		       /* number of connected river tiles  */
};
static struct seadata *oceans;

/* this is used for generator>1 */
#define MAP_NCONT 255


/**************************************************************************
 Wrapper function off the height_map, returns height averages at x,y
**************************************************************************/
static int hmap1(int x, int y, int n)
{
  int nn= 0, hh = 0;

  /* limit to sane distances */
  if( 0 >= n || n >= 5 )
    return hmap(x,map_adjust_y(y));

  square_iterate(x, y, n, xx, yy) {
    nn++, hh+= hmap(xx,yy);
  } square_iterate_end

  return hh/nn;
}

/* Detect highest/lowest (average) neighbouring direction */
static int hslope(int *n, int x, int y, int hi)
{
  int hh,h0,h1,h2,h3, nn,n0;

  /* fudge if called at the poles */
  if( (0 >= y && (++y, n0=0))
   || (map.ysize-1 <= y && (--y, n0=4)) ) {
    if( NULL != n ) *n= n0;
    return hmap(x,y);
  }

  /* 3-tile average of adjacent squares - compute for DXY[0] */
  n0 = 0;
  h3 = hmap(DX1(1)+x,DY1(1)+y);
  h2 = hmap(DX1(0)+x,DY1(0)+y);
  h1 = hmap(DX1(7)+x,DY1(7)+y);
  h0 = h1 + h2 + h3;
  /* wrap to end of DXY vector and walk down */
  for(hh= h0, nn = 8; --nn>0;) {
    hh-= h3, h3= h2, h2= h1;
    hh+= h1= hmap(DX1(nn-1)+x,DY1(nn-1)+y);
    if( (hh<h0) != (hi!=0) ) h0 = hh, n0 = nn;
  }
  if( NULL != n ) *n = n0;

  return ++h0/3;
}

/* Detect lower-altitude mountains (rough) and high level plateaus
 * Hills and mountains at borders of heights, not centres.
*/
static int is_rough(int x, int y)
{
  int i0, i1;

  i0 = hmap(x,map_adjust_y(y));
  i1 = hslope(0,x,y,0);
  if( (i1*4-i0*3) > 0 )
    return 0;       /* not 33% higher than surrounding */
  return 1;
}

static int is_cold(int x, int y)
{
  /* top or bottom 20% (fifth) of map.ysize */
  return ( y*5 < map.ysize || y*5 >= map.ysize*4 );
}

/* Counts tiles of a given type including pseudo rivers within radius n */
static int is_map1(int x, int y, int n, int tt)
{
  int nn = 0;

  /* limit to sane distances */
  if( 0 >= n || n >= 5 )
    return (map_get_terrain(x,y)==tt?1:0);

  if( tt == T_RIVER && terrain_control.river_style == R_AS_SPECIAL) {
    square_iterate(x, y, n, xx, yy) {
      nn+= (map_get_special(xx,yy)&S_RIVER?1:0);
    } square_iterate_end
  }
  else {
    square_iterate(x, y, n, xx, yy) {
      nn+= (map_get_terrain(xx,yy)==tt?1:0);
    } square_iterate_end
  }
  return nn;
}

/* fast approximate true distance without using square roots */
static int map_dist(int p0, int p1)
{
  int i0, i1;

  i0= (p0%map.xsize)-(p1%map.xsize);
  if(i0<0) i0= -i0;
  i1= map.xsize-i0;
  if(i0<i1) i0= i1;
  i1= (p0/map.xsize)-(p1/map.xsize);
  if(i1<0) i1= -i1;
  i0= (i0<i1)?(i1+i0/2):(i0+i1/2);
  return i0;
}

/**************************************************************************
  make_mountains() will convert all squares that are higher than thill to
  mountains and hills. Notice that thill will be adjusted according to
  the map.mountains value, so increase map.mountains and you'll get more
  hills and mountains (and vice versa).
  Add hills/mtns at low-level rough locations, and make 1/3 of highlands
  into plains/deserts/glaciers.
  Note try to balance height and rough, and adjust thill so that there is
  an excess of each for random selection. This spreads the mountains out.
**************************************************************************/
static void make_mountains(int hill)
{
  int mount, rough;
  int j;
  int total=(((map.xsize*map.ysize*map.landpercent)/100)*map.mountains)/100;

  /* Rough is more evenly spread than heights, e.g. near continental shelf */
  rough= ((trough<total?trough:total)*3)/5;
  thill= hill;
  for (j=10;j-->0;) {
    mount=0;
    whole_map_iterate(x, y) {
      if((map_get_terrain(x,y)!=T_OCEAN) && hmap(x,y)>thill && !is_rough(x,y))
	mount++;
    } whole_map_iterate_end;
    thill= (((mount*2)/3+rough<total?9:11)*thill)/10;
    thill= (thill>=maxval)?(maxval-1):thill;
  }

  whole_map_iterate(x, y) {
    /* Note this over selects rough heights somewhat */
    if( (map_get_terrain(x,y)!=T_OCEAN)
     && (((j= is_rough(x, y)) && myrand(trough)<rough)
       || (hmap(x, y)>thill && myrand(100)<66))  ) {
      if ( j ) {
	if (myrand(is_map1(x,y,1,T_OCEAN)*2+3)<2)
	  map_set_terrain(x, y, T_MOUNTAINS);
	else
	  map_set_terrain(x, y, T_HILLS);
      } else if ( myrand(100)<75 ) {
	if ( myrand(maxval-thill)+thill<hmap(x, y))
	  map_set_terrain(x, y, T_MOUNTAINS);
	else
	  map_set_terrain(x, y, T_HILLS);
      } else {
	j= map.ysize*y/100, j=(j>70?j-70:30-j)/2;
	if (myrand(100)<j)
	  map_set_terrain(x, y, T_ARCTIC);
	else if ((j=myrand(100))<35)
	  map_set_terrain(x, y, T_DESERT);
	else if (j<85)
	  map_set_terrain(x, y, T_PLAINS);
      }
    }
  } whole_map_iterate_end;
}

/**************************************************************************
 add arctic and tundra squares in the arctic zone.
 (that is the top 10%, and low 10% of the map)
**************************************************************************/
static void make_polar(void)
{
  int n1,y1,y,x;

  for (y=0;y<map.ysize/10;y++) {
    y1=map.ysize-1 -y;
    n1=(map.ysize/10-y)*1000; /* pseudo-height <- arctic latitude */
    for (x=0;x<map.xsize;x++) {
      if ( y==0
       || (map_get_terrain(x,y)==T_GRASSLAND && myrand(maxval)<hmap(x,y)+n1)
       || (map_get_terrain(x,y)==T_OCEAN && myrand(6)+y<3)) {
	if (myrand(y*50+1) + myrand(y*50+1) < 60)
	  map_set_terrain(x, y, T_ARCTIC);
	else
	  map_set_terrain(x, y, T_TUNDRA);
	
      }
      if ( y==0
       || (map_get_terrain(x,y1)==T_GRASSLAND && myrand(maxval)<hmap(x,y1)+n1)
       || (map_get_terrain(x,y1)==T_OCEAN && myrand(6)+y<3)) {
	if (myrand(y*50+1) + myrand(y*50+1) < 60)
	  map_set_terrain(x, y1, T_ARCTIC);
	else
	  map_set_terrain(x, y1, T_TUNDRA);
      }
    }
  }

  /* only arctic and tundra allowed at the poles (first and last two lines,
     as defined in make_passable() ), to be consistent with generator>1.
     turn every land tile on the second lines that is not arctic into tundra,
     since the first lines have already been set to all arctic above. */
  for (x=0;x<map.xsize;x++) {
    if (map_get_terrain(x, 1)!=T_ARCTIC &&
	map_get_terrain(x, 1)!=T_OCEAN)
      map_set_terrain(x, 1, T_TUNDRA);
    if (map_get_terrain(x, map.ysize-2)!=T_ARCTIC &&
	map_get_terrain(x, map.ysize-2)!=T_OCEAN)
      map_set_terrain(x, map.ysize-2, T_TUNDRA);
  }
}

/**************************************************************************
  recursively generate deserts, use the heights of the map to make a desert
  irregularly shaped, diff is the recursion stopper, and is reduced more if
  desert wants to grow in the y direction, so we end up with "wide" deserts.
**************************************************************************/
static int make_desert(int x, int y, int height, int diff)
{
  int ret=0;
  if( map_get_terrain(x,y)==T_GRASSLAND ) {
    ret++, map_set_terrain(x, y, T_DESERT);
    if (abs(hmap(x, y)-height)<diff) {
      cartesian_adjacent_iterate(x, y, x1, y1) {
	ret+= make_desert(x1, y1, height, ((x != x1)?(diff-100):(diff-300)));
      } cartesian_adjacent_iterate_end;
    }
  }
  return ret;
}

/*************************************************************************
  make_deserts calls make_desert until we have enough deserts actually
  there is no map setting for how many we want, what happends is that
  we choose a random coordinate between 20 and 30 degrees north and south
  (deserts tend to be between 15 and 35, make_desert will expand them) and
  if it's a grassland square we call make_desert with this coordinate, we
  try this 500 times for each region: north and south.
**************************************************************************/
static void make_deserts(void)
{
  int x,y, n, nn= map.ysize*10;
  int deserts=(((map.xsize*map.ysize*map.landpercent)/100)*map.deserts)/100;

  /* temperate latitudes: 20-40, 60-80 centered on 30,70% ysize */
  for( n= deserts*8; deserts>0 && n-->0; ) {
    y= (myrand(nn)+myrand(nn)+map.ysize*60)/100;
    x= myrand(map.xsize);
    if (map_get_terrain(x, y)==T_GRASSLAND) {
      deserts-= make_desert(x,y, hmap(x, y), 1000);
    }
    y= (myrand(nn)+myrand(nn)+nn+nn)/100;
    x= myrand(map.xsize);
    if (map_get_terrain(x, y)==T_GRASSLAND) {
      deserts-= make_desert(x,y, hmap(x, y), 1000);
    }
  }
}

/**************************************************************************
  a recursive function that adds forest to the current location and tries
  to spread out to the neighbours, it's called from make_forests until
  enough forest has been planted. diff is again the block function.
  if we're close to equator it will with 50% chance generate jungle instead
  decrease slightly in the x direction to reflect east west rainfall loss
**************************************************************************/
static void make_forest(int x, int y, int height, int diff)
{
  int nn;

  if (y==0 || y==map.ysize-1)
    return;

  if (map_get_terrain(x, y)==T_GRASSLAND) {
    nn= y*100 / map.ysize;
    nn= nn > 50 ? 100-nn : nn;
    nn= myrand(nn)+myrand(nn);
    map_set_terrain(x, y,
      (nn > 40 ? T_JUNGLE : (nn > 5 ? T_FOREST : T_SWAMP)));
    if (abs(hmap(x, y)-height)<diff) {
      cartesian_adjacent_iterate(x, y, x1, y1) {
	if (myrand(10)<4)
	  make_forest(x1, y1, height, ((x != x1)?(diff-600):(diff-400)));
      } cartesian_adjacent_iterate_end;
    }
    forests++;
  }
}

/**************************************************************************
  makeforest calls make_forest with random grassland locations until there
  has been made enough forests. (the map.forestsize value controls this)
**************************************************************************/
static void make_forests(void)
{
  int x,y, forever;
  int forestsize;
  forever=(map.xsize*map.ysize*map.landpercent)/50;
  forestsize=(forever*map.forestsize)/200;
  forests=0;
  do {
    /* boreal and equatorial bands */
    y=(myrand(map.ysize*2)+(myrand(3)*3+1)*map.ysize)/10;
    x=myrand(map.xsize);
    make_forest(x,y, hmap(x, y), 2500);
  } while ((forever-->0) && (forests<forestsize));
}

/**************************************************************************
  swamps are placed on low lying locations, that will typically be close to
  the shoreline. They're put at random (where there is lowlying grassland)
  and with 50% chance each of it's neighbour squares will be converted to
  swamp as well.
  Note: some swamps are really lakes.
**************************************************************************/
static void make_swamps(void)
{
  int x,y, t0, tt, swamps, forever;
  forever= (map.xsize*map.ysize*map.landpercent)/50;
  swamps= (forever*map.swampsize)/200;
  while (swamps > 0 && forever-- > 0) {
    y= myrand(map.ysize-4)+2;
    x= myrand(map.xsize);
    tt= map_get_terrain(x, y);
    t0= T_SWAMP;
    if( (tt == T_GRASSLAND && hmap(x, y)<thill)
     || (tt != T_OCEAN && (is_map1(x, y, 1, T_OCEAN)>2 || (t0= T_OCEAN))) ) {
      swamps--, map_set_terrain(x, y, t0);
      adjc_iterate(x, y, x1, y1) {
	tt= map_get_terrain(x1, y1);
	if( (tt == T_GRASSLAND && myrand(10)<5)
	 || (tt == T_JUNGLE && myrand(10)<3)
	 || (tt == T_FOREST && myrand(10)<2) )
	  swamps--, map_set_terrain(x1, y1, t0);
      } adjc_iterate_end;
    }
  }
}

/*************************************************************************
 this recursive function tries to make some decent rivers, that run towards
 the ocean, it does this by following a descending path on the map spiced
 with a bit of chance, if it fails to reach the ocean it rolls back.
**************************************************************************/
static int make_river(int x, int y, int force)
{
  static int ix[]= { 0, 250, 0, 300, 0, 250, 0, 200 };
  static int rlength= 0, olength= 0, ovalue= 0;
  enum tile_terrain_type tt, tx, ty, tz;
  int res, min0, min1, osave;
  int nr, no, m0, m1, xx, yy;

  /* ignore poles as well as off-map */
  if (y <= 0 || y >= map.ysize-1)
    return 0;

  /* oops checks */
  tt= map_get_terrain(x, y);
  if (tt>=16)
    return 0;
  if (tt == T_RIVER || (map_get_special(x, y)&S_RIVER))
    return 0;

  /* (Re)set ovalue to current tile, unless we have passed a lake,
   *   then take the minimum positive value.
  */
  no= map_get_ocean(x, y);
  if( olength == rlength || (no > 0 && (no < ovalue || ovalue == 0)) )
    ovalue= no;

  /* collect surrounding river/ocean stats
   * nr    - number of adjacent river squares
   * no    - number of adjacent ocean squares (i.e. THE ocean)
   * m0,xx - number of adjacent sea squares, and a direction
   * m1,yy - number of diagonal sea squares, and a direction
   * TODO: collect the highest and lowest ocean value+direction
  */
  res= -1, xx= -1, yy= -1, m0= 0, m1= 0, no= 0, nr= 0;
  cartesian_adjacent_iterate(x, y, x1, y1) {

    /* diagonals - cheat diagonal is at -1 */
    no+= (map_get_ocean(DX1(_nn-1)+x,DY1(_nn-1)+y)==1?1:0);
    ty= (map_get_terrain(DX1(_nn-1)+x,DY1(_nn-1)+y)&15);
    m1+= (ty==T_OCEAN?(yy=_nn-1,1):0);

    /* cartesian */
    no+= (map_get_ocean(x1,y1)==1?1:0);
    tx= (map_get_terrain(x1,y1)&15);
    m0+= (tx==T_OCEAN?(res=xx, xx=_nn, 1):0);

    if( tx==T_RIVER || map_get_special(x1,y1)&S_RIVER ) {
       /* do not end next to (i.e. make river fork in) these terrain */
       switch(tx) {
       case T_ARCTIC:
       case T_MOUNTAINS:
	 nr++;
       default:
	 break;
       }
       res= xx, xx=_nn, nr++;
    }
  } cartesian_adjacent_iterate_end

  /* terminate on THE ocean, not a tiny lake
   * Note: if we continue, lakes have better connectivity (+diagonals)
   *       than rivers
   */
  if ( tt == T_OCEAN ) {
    if( ovalue == 1)
      return 1;    /* hit THE ocean */
    if( (m0+= m1) > 1 )
      return 1;   /* big lake, not a pseudo-river - too many choices */
    if( xx == -1 ) xx= yy;
    else
    if( res == -1 ) res= yy;
  }

  if ( !force ) {
    /* did we enter a tile bordered by river(s) and/or ocean? */
    switch(nr+(m0>0?1:0)) {
    default:
      /* no double outflows */
      return 0;
    case 2:
      /* join up inland river/lake */
      tx= map_get_ocean(DX1(xx)+x, DY1(xx)+y);
      ty= map_get_ocean(DX1(res)+x, DY1(res)+y);
      if( tx == ty
       || tx == ovalue
       || ty == ovalue
       || m0 == 0
       || (m0+nr) != 2
       || (m0+nr)+ty == no )
	return 0;
      m0= 0; /* and FALL THROUGH */
    case 1:
      tx= map_get_ocean(DX1(xx)+x, DY1(xx)+y);
      if(m0>0) {
	/* can follow a single sea, it is probably a lake */
	if( m0 == 1
	 && no == 0
	 && (rlength==olength || myrand(olength)<2) )
	  break;

	/* join *exactly* two different source lakes */
	/* There is a hole here. 3-6 adjacent oceans can also be part of
	   2 distinct seas, and there is no check for "different" in these
	   cases, allowing double outlet through a short river. Because it
	   is difficult to recognize which 2 are different, and the 3-6
	   cases have value on its own most of the time, we let this go. */
	if( m0 == 2 ) {
	  ty= map_get_ocean(DX1(res)+x, DY1(res)+y);
	  if( tx == ty
	   || tx == ovalue
	   || ty == ovalue )
	    return 0;
	}
	else /* bias against short coastal creeks */
	if( myrand(3)+(tt==T_OCEAN?1:0) >= rlength )
	  return 0;
      }
      /* join up! */
      if (tt == T_OCEAN)
	;
      else if (terrain_control.river_style == R_AS_TERRAIN)
	map_set_terrain(x, y, T_RIVER);
      else if (terrain_control.river_style == R_AS_SPECIAL)
	map_set_special(x, y, S_RIVER);
      else
	return 0;  /* paranoia? */
      map_set_ocean(x, y, 0);
      assign_ocean_flood1(x, y, -1, tx);
      ovalue= map_get_ocean(x, y);
      return 1;
    case 0:
      /* keep on paddling */
      break;
    }
  }

  /* probe two directions for next possible river tile and recurse */
  /* temporarily increment the counts and block the current square */
  osave = olength, olength++, rlength++;
  map_set_terrain(x, y, tt|16);

  if( m0 == 1 ) {
    /* no choice - step into the lake (or fjord) - try twice */
    m0= m1= xx;
    olength = 0;
  } else {
    /* extend the river - slight bias to the east (ix), 2 tries,
     * bias towards entering any costal square or towards ocean,
     * bias against entering arctic, mountains and hills
    */
    min0= hmap1(x, y, 1)+(tt == T_OCEAN ? 2 : 1)*ix[1];
    m0= m1= -1;
    cartesian_adjacent_iterate(x, y, x1, y1) {
      if( (tx= map_get_terrain(x1, y1)) < 16
       && (tx != T_OCEAN)
       && (tx != T_ARCTIC)
       && (tx != T_MOUNTAINS || myrand(16) < 1)
       && (tx != T_HILLS || myrand(8) < 1)
       && (tx != T_DESERT || myrand(4) < 1)
       && (res= hmap1(x1,y1,1))-myrand
	  (is_map1(x1,y1,1,T_OCEAN)?3:
	  (is_map1(x1,y1,2,T_OCEAN)?2:1)*ix[_nn]) < min0) {
	min1= min0, m1= m0;
	min0= res,  m0= _nn;
      }
    } cartesian_adjacent_iterate_end
  }
  for(res=0; res == 0 && m0 != -1; m0= m1, m1= -1) {
    /* block orthogonal land tiles - cartesian (+/-2) or diagonal (+/-1) */
    /* TODO: diagonal sea walk does not block all looping tail biters */
    /*       +/-3 block would help, but probably need +/-2 as well    */
    nr= (m0&1)+1;
    xx= DX1(m0-nr)+x, yy= DY1(m0-nr)+y;
    tx= map_get_terrain(xx, yy);
    if (tx != T_OCEAN)
      map_set_terrain(xx, yy, tx|16);
    xx= DX1(m0+nr)+x, yy= DY1(m0+nr)+y;
    ty= map_get_terrain(xx, yy);
    if (ty != T_OCEAN)
      map_set_terrain(xx, yy, ty|16);
    xx= DX1(m0+4)+x, yy= DY1(m0+4)+y; /* and looping tail biters, grrr ,,, */
    tz= map_get_terrain(xx, yy);
    if (tz != T_OCEAN)
      map_set_terrain(xx, yy, tz|16);

    /* step ... */
    res= make_river(DX1(m0)+x,DY1(m0)+y,0);

    /* unblock neighbours */
    map_set_terrain(xx, yy, tz);
    xx= DX1(m0+nr)+x, yy= DY1(m0+nr)+y;
    map_set_terrain(xx, yy, ty);
    xx= DX1(m0-nr)+x, yy= DY1(m0-nr)+y;
    map_set_terrain(xx, yy, tx);
  }
  map_set_terrain(x, y, tt);
  olength=osave, rlength--;

  /* back from recurse, mark this tile if successful and return result */
  if (res || force || tt==T_OCEAN) {
    res++;
    if (tt==T_OCEAN)
      ;
    else if (terrain_control.river_style==R_AS_TERRAIN) {
      map_set_terrain(x, y, T_RIVER);
    }
    else if (terrain_control.river_style==R_AS_SPECIAL) {
      map_set_special(x, y, S_RIVER);
    }
    map_set_ocean(x, y, 0);
    assign_ocean_flood1(x, y, -1, ovalue);
    ovalue= map_get_ocean(x, y);
  }
  return res;
}

/**************************************************************************
  calls make_river until we have enough river tiles (map.riverlength),
  or forever patience runs out.
  Finally, try one last time to connect isolated inland watersheds.
**************************************************************************/
static void make_rivers(void)
{
  enum tile_terrain_type tx, tt;
  int m0, m1, no, nr;
  int x, y, xy, *yx, rivers, forever, k=0;

  assign_ocean_numbers();

  /* collect a randomized list of eligible tiles */
  xy= map.ysize*map.xsize;
  yx= fc_malloc(sizeof(int)*xy);
  block_iterate(x, y, 0, 1, map.xsize, map.ysize-2) {
    if (map_get_terrain(x,y) != T_OCEAN)
      tx= myrand(k+1), yx[k++]= yx[tx], yx[tx]= _xy;
  } block_iterate_end;

  /* Assume map.riverlength is an absolute number of squares */
  /* Ignore small values that might be a misplaced percent   */
  /* TODO: get concensus to reverse this assumption and use percent */
  rivers= (map.xsize*map.ysize*map.landpercent*20)/10000;
  if( map.riverlength==0 || map.riverlength>rivers/2 )
    rivers= map.riverlength;

  /* River generation is quite picky, try roughly 4 times per tile */
  forever= k*4;
  while (rivers > 0 && forever-- > 0) {
    x= yx[myrand(k)], y= x/map.xsize, x= x%map.xsize;
    tt= map_get_terrain(x, y);

    /* collect surrounding river/ocean stats
     * nr - number of adjacent river squares
     * m0 - number of adjacent ocean squares
     * m1 - number of diagonal ocean squares
    */
    m0= 0, m1= 0, nr= 0, no= 0;
    cartesian_adjacent_iterate(x, y, x1, y1) {

      /* diagonals - cheat diagonal is at -1 */
      no+= (map_get_ocean(DX1(_nn-1)+x,DY1(_nn-1)+y)==1?1:0);
      tx= map_get_terrain(DX1(_nn-1)+x,DY1(_nn-1)+y);
      m1+= (tx == T_OCEAN?1:0);

      /* cartesian */
      no+= (map_get_ocean(x1,y1) == 1?1:0);
      tx= map_get_terrain(x1,y1);
      m0+= (tx == T_OCEAN?1:0);

      if( tx == T_RIVER || map_get_special(x1,y1)&S_RIVER ) {
	/* do not end next to (i.e. make river fork in) these terrain */
	switch(tx) {
	case T_MOUNTAINS:
	case T_ARCTIC:
	  nr++;
        default:
	  break;
	}
	nr++;
      }
    } cartesian_adjacent_iterate_end

    if( (m0+nr) == 2
     && (m0+nr)+m1 != no
     && tt != T_OCEAN
     && !(tt == T_RIVER || map_get_special(x, y)&S_RIVER) )
      ;  /* join lakes */
    else /* when not to start here */
    if( hmap(x,y) < myrand((maxval*60)/100)  /* bias to higher elevations */
     || tt == T_OCEAN
     || tt == T_DESERT
     || (tt == T_MOUNTAINS && myrand(16) > 0)
     || (tt == T_HILLS && myrand(8) > 0)
     || (tt == T_TUNDRA && myrand(4) > 0)
     || (tt == T_RIVER || map_get_special(x, y)&S_RIVER)
     || (((m0+= m1) > 0 ? 1 : 0)+nr > 1)     /* bad water connectivity */
     || (m0 > 0 && myrand(4) > (m0-no))      /*   quickie tests ...    */
     || (is_map1(x,y,2,T_OCEAN)+is_map1(x,y,2,T_RIVER) > 3 && myrand(3) > 0)
    )
      continue;

    rivers-=make_river(x,y,0);
  }

  /* for each *inland* watershed
   *   list tiles
   *   for each tile
   *     walk adjacent2 (cartesian>knights>diagonal preference)
   *       if is_different_watershed then
   *         join up via intervening adjacent tile
  */
  setup_seadata();
  for( k= map_num_oceans()+1; --k > 1; ) {
    /* retest for the case where we just joined */
    if( (m0= map_get_ocean(oceans[k].x, oceans[k].y)) == 1 )
      continue;
    yx[0]= xy-2, yx[1]= 0;
    rivers= list_ocean_flood(yx+1, oceans[k].x, oceans[k].y);
    for( rivers++; --rivers > 0; ) {
      y= yx[rivers+1]/map.xsize, x= yx[rivers+1]%map.xsize;
      tx= T_OCEAN, tt= map_get_terrain(x, y);

      /* find nearby inland waterways */
      no= -1;
      cartesian_adjacent_iterate2(x, y, x1, y1) {
	if( (m1= map_get_ocean(x1,y1)) > 0 && m1 != m0 ) {
	  tx=map_get_terrain(x1,y1), no= _nn/2;
	  if( map_get_ocean(DX1(no-1)+x, DY1(no-1)+y) == 0
	   && map_get_ocean(DX1(no+1)+x, DY1(no+1)+y) == 0 ) 
	    break;
	  no= -1;
	}
      } cartesian_adjacent_iterate2_end;
      if( no == -1 ) {
	cartesian_knights_iterate2(x, y, x1, y1) {
	  if( (tt == T_OCEAN || map_get_terrain(x1,y1) == T_OCEAN)
	   && (m1= map_get_ocean(x1,y1)) > 0
	   &&  m1 != m0 ) {
	    tx= T_OCEAN, no= (tt==T_OCEAN ? ((_nn+1)&-3) : ((_nn&-3)+2))/2;
	    if( map_get_ocean(DX1(no-1)+x, DY1(no-1)+y) == 0
	     && map_get_ocean(DX1(no+1)+x, DY1(no+1)+y) == 0 ) 
	      break;
	    no= -1;
	  }
	} cartesian_knights_iterate2_end;
      }
      if( no == -1 ) {
	cartesian_diagonal_iterate2(x, y, x1, y1) {
	  if( (tt == T_OCEAN && map_get_terrain(x1,y1) == T_OCEAN)
	   && (m1= map_get_ocean(x1,y1)) > 0
	   &&  m1 != m0 ) {
	    tx= T_OCEAN, no= _nn/2;
	    if( map_get_ocean(DX1(no-1)+x, DY1(no-1)+y) == 0
	     && map_get_ocean(DX1(no+1)+x, DY1(no+1)+y) == 0 ) 
	      break;
	    no= -1;
	  }
	} cartesian_diagonal_iterate2_end;
      }
      if( no != -1 ) {
	/* hookup via intervening adjacent tile */
	x+= DX1(no), y+= DY1(no);
	if( tx == T_OCEAN
	 || tx == T_ARCTIC
	 || tx == T_MOUNTAINS ) {
	  map_clear_special(x, y, S_RIVER);
	  map_set_terrain(x, y, T_OCEAN);
	  map_set_continent(x, y, 0);
	}
	else if (terrain_control.river_style==R_AS_TERRAIN)
	  map_set_terrain(x, y, T_RIVER);
	else if (terrain_control.river_style==R_AS_SPECIAL)
	  map_set_special(x, y, S_RIVER);

	map_set_ocean(x, y, 0);
	assign_ocean_flood1(x, y, -1, m0);
	m0= map_get_ocean(x, y);
      }
    }
  }
  free(yx);
}

/**************************************************************************
  make_plains converts ~50% of the remaining grassland to plains based on
  height and latitude (plains are more likely at desert height/latitudes).
**************************************************************************/
static void make_plains(void)
{
  int x,y,y0,y1,nn, ng,np;

  do {
    ng= np= 0;
    for (y=1;y<map.ysize/2;y++) {
      y1 = map.ysize-1 -y;
      y0 = (y*10000)/map.ysize;   /* latitude to pseudo altitude */
      nn = (y0 < 2000) ? (4000-y0):y0;
      for (x=0;x<map.xsize;x++) {
	if( map_get_terrain(x, y)==T_GRASSLAND
	 && (myrand(hmap(x,y)+7000)>tocean+nn))
	  map_set_terrain(x, y, (myrand(25)>y0?T_TUNDRA:T_PLAINS));
	ng+= map_get_terrain(x,y)==T_GRASSLAND?1:0;
	np+= map_get_terrain(x,y)==T_PLAINS?1:0;

	if( map_get_terrain(x, y1)==T_GRASSLAND
	 && (myrand(hmap(x,y1)+7000)>tocean+nn))
	  map_set_terrain(x, y1, (myrand(25)>y0?T_TUNDRA:T_PLAINS));
	ng+= map_get_terrain(x,y1)==T_GRASSLAND?1:0;
	np+= map_get_terrain(x,y1)==T_PLAINS?1:0;
      }
    }
  } while (ng*8>np*10);
}

/**************************************************************************
  we want the map to be sailable east-west at least at north and south pole
  and make it a bit jagged at the edge as well.
  So this procedure converts 60% of the second line and the second last line
  to ocean, 40% of the 4th and 4th last line to ocean, and 50% of the 3rd
  and 3rd last line to ocean (100% if the 4th was not ocean).
**************************************************************************/
static void make_passable(void)
{
  int x;

  for (x=0;x<map.xsize;x++) {
    if (myrand(100)<40) map_set_terrain(x,1,T_OCEAN);
    if (myrand(100)<60) map_set_terrain(x,3,T_OCEAN);
    if ( map_get_terrain(x-1,3) != T_OCEAN
      || map_get_terrain(x  ,3) != T_OCEAN
      || map_get_terrain(x+1,3) != T_OCEAN
      || myrand(100)<50)
      map_set_terrain(x,2,T_OCEAN);
    if (myrand(100)<40) map_set_terrain(x,map.ysize-2,T_OCEAN);
    if (myrand(100)<60) map_set_terrain(x,map.ysize-4,T_OCEAN);
    if ( map_get_terrain(x-1,map.ysize-4) != T_OCEAN
      || map_get_terrain(x  ,map.ysize-4) != T_OCEAN
      || map_get_terrain(x+1,map.ysize-4) != T_OCEAN
      || myrand(100)<50)
      map_set_terrain(x,map.ysize-3,T_OCEAN);
  }
}

/**************************************************************************
  we don't want huge areas of monoterrain like grass/plains,
  so we put in a hill here and there, where it gets too 'clean'.
  Actually we can do a little better than that.
**************************************************************************/
static void make_fair(void)
{
  int ii, nn, t0, tt, x,y, xy,*yx;

  /* randomize hits to avoid NW->SE waves - exclude poles */
  xy = (map.ysize-4)*map.xsize;
  yx = (int *)fc_malloc(xy*sizeof(int));
  for( yx[0]= 0, y=1; y<xy; y++)
    x= myrand(y+1), yx[y]= yx[x], yx[x]= y;

  while (xy-- > 0) {
    x = yx[xy] % map.xsize;
    y = yx[xy] / map.xsize + 2;

    t0=map_get_terrain(x,y);
    switch(t0) {
    case T_MOUNTAINS:
      if (is_at_coast(x,y) && (hslope(&nn,x,y,0),nn=nn+4)) {
	if( myrand(5)<4
	 && ((tt= map_get_terrain(DX1(nn)+x,DY1(nn)+y)) != T_OCEAN)) {
	  /* hack - increase coastal plain swap t0 one square inland */
	  map_set_terrain(DX1(nn)+x,DY1(nn)+y,t0);
	  map_set_terrain(x,y,t0=tt);
	}
	nn=2;
      }
      else
	nn=5;

      if (is_map1(x,y,2,t0)+is_map1(x,y,1,T_HILLS)>nn) {
	for(ii=2; ii-->0; ) {
	  if ((myrand(4)<3)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(5)<1 ? T_ARCTIC : (myrand(3)<1 ? T_PLAINS : T_FOREST);
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    case T_HILLS:
      if (is_at_coast(x,y) && (hslope(&nn,x,y,0),nn=nn+4)) {
	if( myrand(3)<2
	 && ((tt= map_get_terrain(DX1(nn)+x,DY1(nn)+y)) != T_OCEAN)) {
	  /* hack - increase coastal plain swap t0 one square inland */
	  map_set_terrain(DX1(nn)+x,DY1(nn)+y,t0);
	  map_set_terrain(x,y,t0=tt);
	}
	nn=3;
      }
      else
	nn=6;

      if (is_map1(x,y,2,t0)+is_map1(x,y,1,T_MOUNTAINS)>nn) {
	for(ii=3; ii-->0; ) {
	  if ((myrand(4)<2)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(3)<1 ?  T_PLAINS: T_FOREST;
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    case T_ARCTIC:
    case T_TUNDRA:
    case T_DESERT:
      if (is_map1(x,y,2,t0)>20) {
	for(ii=2; ii-->0; ) {
	  if ((myrand(3)<2)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(3)<1 ? T_MOUNTAINS : T_HILLS;
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    case T_JUNGLE:
    case T_FOREST:
      if (is_map1(x,y,2,T_JUNGLE)+is_map1(x,y,2,T_FOREST)>20) {
	for(ii=3; ii-->0; ) {
	  if ((myrand(3)<2)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(3)<1 ? T_SWAMP : (myrand(3)<2 ? T_GRASSLAND : T_HILLS);
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    case T_SWAMP:
      if (is_map1(x,y,2,t0)>20) {
	for(ii=3; ii-->0; ) {
	  if ((myrand(3)<2)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(3)<1 ? T_FOREST : (myrand(3)<2 ? T_GRASSLAND : T_HILLS);
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    case T_PLAINS:
    case T_GRASSLAND:
      if (is_map1(x,y,3,T_PLAINS)+is_map1(x,y,3,T_GRASSLAND)>40) {
	for(ii=6; ii-->0; ) {
	  if ((myrand(3)<2)
	   && ((nn = myrand(8)), map_get_terrain(DX1(nn)+x,DY1(nn)+y))==t0) {
	    tt = myrand(3)<1 ? T_FOREST : T_HILLS;
	    map_set_terrain(DX1(nn)+x,DY1(nn)+y, tt);
	  }
	}
      }
      break;
    }
  }
  free(yx);
}

/**************************************************************************
  Returns 1 if this is a 1x1 island (weighted by diagonal accessibility)
**************************************************************************/
static int is_tiny_island(int x, int y)
{
  static int ix[] = { 1, 16, 2, 32, 4, 64, 8, 128 };
  int nn = 0;
  if (map_get_terrain(x,y)==T_OCEAN) return 0;

  /* bias against small archipelagos outside coastal waters */
  if (is_map1(x,y,2, T_OCEAN)+myrand(4)>24)
     return 1;

  cartesian_adjacent_iterate(x,y,x1,y1) {
    if (map_get_terrain(x1,y1)!=T_OCEAN) return 0;
  } cartesian_adjacent_iterate_end

  cartesian_diagonal_iterate(x,y,x1,y1) {
    if (map_get_terrain(x1,y1)!=T_OCEAN) nn+=ix[_nn];
  } cartesian_diagonal_iterate_end

  /* some patterns are more aesthetically pleasing than others */
  switch(nn) {
  case 1:
  case 2:
  case 4:
  case 8:
    return (myrand(100) < 66 ? 1:0);
  case 3:
  case 12:
    return (myrand(100) < 75 ? 1:0);
  case 5:
  case 6:
  case 9:
  case 10:
    return (myrand(100) < 50 ? 1:0);
  case 7:
  case 11:
  case 13:
  case 14:
    return (myrand(100) < 90 ? 1:0);
  case 15:
  case 0:
    return 1;
  }
  return 0;
}

/**************************************************************************
  Removes all 1x1 islands (sets them to ocean).
**************************************************************************/
static void remove_tiny_islands(void)
{
  block_iterate(x, y, 0, 3, map.xsize, map.ysize-6) {
    if (is_tiny_island(x, y)) {
      map_set_terrain(x, y, T_OCEAN);
      map_clear_special(x, y, S_RIVER);
      map_set_continent(x, y, 0);
    }
  } block_iterate_end;
}

/**************************************************************************
  Returns 1 if this is a 1x1 lake (weighted by diagonal accessibility)
**************************************************************************/
static int is_tiny_lake(int x, int y)
{
  static int ix[] = { 16, 1, 32, 2, 64, 4, 128, 8 };
  int nr= 0, nn = 0;

  if (map_get_terrain(x,y) != T_OCEAN) return 0;
  if (map_get_ocean(x,y) == 1) return 0;

  /* Don't try to unravel river flows */
  cartesian_adjacent_iterate(x, y, x1, y1) {
    nn+= (map_get_terrain(x1, y1) == T_OCEAN?ix[_nn]:0);
    nr+= (map_get_terrain(x1, y1) == T_RIVER
       || map_get_special(x1, y1)&S_RIVER)?1:0;
  } cartesian_adjacent_iterate_end
  if( nr > 0 )
    return 0;

  cartesian_diagonal_iterate(x, y, x1, y1) {
    nn+= (map_get_terrain(x1, y1) == T_OCEAN?ix[_nn]:0);
  } cartesian_adjacent_iterate_end

  /* bias against small (potentially) landlocked seas */
  if (is_map1(x,y,2, T_OCEAN)+myrand(4)<4)
     return 1;

  /* some patterns are more aesthetically pleasing than others */
  switch(nn) {
  case 1:
  case 2:
  case 4:
  case 8:
    return (myrand(100) < 66 ? 1:0);
  case 3:
  case 5:
  case 6:
  case 9:
  case 10:
  case 12:
    return (myrand(100) < 50 ? 1:0);
  case 7:
  case 11:
  case 13:
  case 14:
    return (myrand(100) < 33 ? 1:0);
  case 15:
    return (myrand(100) < 90 ? 1:0);
  case 0:
  case 240:
    return 1;
  }
  return 0;
}

/**************************************************************************
  Removes all 1x1 lakes (sets them to swamp).
**************************************************************************/
static void remove_tiny_lakes(void)
{
  /* Be careful not to break a polar passage */
  block_iterate(x, y, 0, 4, map.xsize, map.ysize-8) {
    if (is_tiny_lake(x,y))
      map_set_terrain(x,y, T_SWAMP);
  } block_iterate_end;
}

/**************************************************************************
  make land simply does it all based on a generated heightmap
  1) with map.landpercent it generates a ocean/grassland map
  2) it then calls the above functions to generate the different terrains
**************************************************************************/
static void make_land(void)
{
  int count, forever=50;
  int total=(map.xsize*map.ysize*map.landpercent)/100;

  /* Adjust sealevel to approximate landpercent */
  tocean=(maxval*map.landpercent)/100;
  do {
    count=0;
    whole_map_iterate(x, y) {
      if (hmap(x, y)<tocean)
	map_set_terrain(x, y, T_OCEAN);
      else {
	map_set_terrain(x, y, T_GRASSLAND);
	count++;
      }
    } whole_map_iterate_end
    tocean= (((count>total)?11:9)*tocean)/10;
  } while (forever-->0 && abs(total-count)*20>total);

  /* Compute working stats */
  trough= 0;
  whole_map_iterate(x, y) {
    trough+= (map_get_terrain(x,y)==T_OCEAN)?0:is_rough(x,y);
  } whole_map_iterate_end

  make_mountains((maxval*66)/100);
  make_deserts();
  make_forests();
  make_swamps();
  make_polar();
  make_fair();
  make_passable();
  make_plains();
  make_rivers();
  remove_tiny_islands();
  remove_tiny_lakes();
  assign_continent_numbers();
  setup_seadata();
  map_check();
}

/**************************************************************************
  this function spreads out huts on the map, a position can be used for a
  hut if there isn't another hut close and if it's not on the ocean.
**************************************************************************/
static void make_huts(int number)
{
  int x,y;
  int count=map.ysize*map.xsize;
  count= (number*count)/1000;
  while (number>0 && count-->0) {
    x=myrand(map.xsize);
    y=myrand(map.ysize-2)+1;
    if (map_get_terrain(x, y)!=T_OCEAN &&
	( map_get_terrain(x, y)!=T_ARCTIC || myrand(10)<3 )
	) {
      if (!is_hut_close(x,y)) {
	number--;
	map_get_tile(x,y)->special|=S_HUT;
	/* Don't add to islands[].goodies because islands[] not
	   setup at this point, except for generator>1, but they
	   have pre-set starters anyway. */
      }
    }
  }
}

static void add_specials(int prob)
{
  enum tile_terrain_type ttype;
  int x,y, xy, *yx;

  /* randomize the tile order */
  xy = (map.ysize-2)*map.xsize;
  yx = (int *)fc_malloc(xy*sizeof(int));
  for( yx[0]= 0, y=1; y<xy; y++)
    x= myrand(y+1), yx[y]= yx[x], yx[x]= y;

  while (xy-- > 0) {
    x = yx[xy] % map.xsize;
    y = yx[xy] / map.xsize + 1;
    ttype = map_get_terrain(x, y);
    if ( ttype!=T_OCEAN || is_coastline(x,y) ) {
      if (myrand(1000) < (ttype==T_GRASSLAND?prob*4/3:prob) ) {
	if (ttype==T_GRASSLAND || !is_special_close(x,y)) {
	  if (tile_types[ttype].special_1_name[0] &&
	    (!(tile_types[ttype].special_2_name[0]) || (myrand(100)<50))) {
	    map_get_tile(x,y)->special|=S_SPECIAL_1;
	  }
	  else if (tile_types[ttype].special_2_name[0]) {
	    map_get_tile(x,y)->special|=S_SPECIAL_2;
	  }
	}
      }
    }
  }
  free(yx);
  map.have_specials = 1;
}

/**************************************************************************
 Number this tile and recursive adjacent tiles with unassigned number
 to specified continent number, by flood-fill algorithm.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int assign_continent_flood(int x, int y, int nr)
{
  return assign_continent_flood1(x, y, 0, nr);
}

/**************************************************************************
 Number this tile and recursive adjacent tiles with given number to
 specified continent number, by flood-fill algorithm, i.e. join.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int assign_continent_flood1(int x, int y, int n0, int nr)
{
  int  ret= 0;

  if( (y>=0 && y<map.ysize)
   && (map_get_continent(x, y)==n0)
   && (map_get_terrain(x, y)!=T_OCEAN) ) {
    ret++, map_set_continent(x, y, nr);
    adjc_iterate(x, y, x1, y1) {
      ret+= assign_continent_flood(x1, y1, nr);
    } adjc_iterate_end;
  }
  return ret;
}

/**************************************************************************
 List this tile and recursively adjacent continent tiles with number "n0"
**************************************************************************/
static int list_continent_flood1(int *list, int x, int y, int n0)
{
  if( map_get_continent(x, y) == n0 ) {
    /* block this square */
    map_set_continent(x, y, -n0);
    adjc_iterate(x, y, x1, y1) {
      list_continent_flood1(list, x1, y1, n0);
    } adjc_iterate_end;
    list[++list[0]]= y*map.xsize+x, map_set_continent(x, y, n0);
  }
  return list[0];
}

/**************************************************************************
 Assign continent numbers to all tiles.
 Numbers 1 and 2 are reserved for polar continents if
 map.generator != 0; otherwise are not special.
 Also sets map.num_continents (note 0 is ocean, and continents
 have numbers 1 to map.num_continents _inclusive_).
 Note this is not used by generators>1 at map creation
 time, as these assign their own continent numbers.
**************************************************************************/
void assign_continent_numbers(void)
{
  int isle = 0;

  whole_map_iterate(x, y) {
      map_set_continent(x, y, 0);
  } whole_map_iterate_end;

  if (map.generator != 0) {
    assign_continent_flood(0, 0, ++isle);
    assign_continent_flood(0, map.ysize-1, ++isle);
  }

  whole_map_iterate(x, y) {
    if (!map_get_continent(x, y) && map_get_terrain(x, y)!=T_OCEAN) {
      assign_continent_flood(x, y, ++isle);
    }
  } whole_map_iterate_end;

  map.num_continents = isle;
  freelog(LOG_VERBOSE, "Map has %d continents", map.num_continents);
}

/**************************************************************************
 Number this tile and recursive adjacent water tiles with unasssigned
   number to specified ocean number, by flood-fill algorithm.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int assign_ocean_flood(int x, int y, int nr)
{
  return assign_ocean_flood1(x, y, 0, nr);
}

/**************************************************************************
 Number this tile and recursively adjacent water tiles with number "n0"
   to ocean number "nr", by flood-fill algorithm (i.e. join watersheds).
 Returns the number of tiles successfully assigned number "nr".
 If n0 is negative, sets all positive tiles to the lowest number found.
 In this case the return totals are meaningless due to double counting.
**************************************************************************/
static int assign_ocean_flood1(int x, int y, int n0, int nr)
{
  int  tx, mm, ret= 0;

  assert(n0 != nr);

  /* sanity check for valid map, water tile */
  if( (y>=0 && y<map.ysize)
   && ( ((tx= map_get_terrain(x, y)&15) == T_OCEAN)
     || (tx == T_RIVER)
     || (map_get_special(x, y)&S_RIVER) )) {
    /* loop: "if" in the normal case, repeat if changed for n0 == -1 */
    /* TODO: should we clear the current tile to always force one cycle? */
    while( (mm= map_get_ocean(x, y),n0) >= 0
	?(mm == n0):(mm != nr && (mm > nr || mm <= 0 || (nr= mm))) ) {
      /* update this square */
      ret++, map_set_ocean(x, y, nr);
      /* update river accessible adjacent */
      cartesian_adjacent_iterate(x, y, x1, y1) {
	ret+= assign_ocean_flood1(x1, y1, n0, nr);
	nr= ((mm= map_get_ocean(x1, y1)) > 0 ? mm : nr);
      } cartesian_adjacent_iterate_end;
      if(tx == T_OCEAN) {
	/* plus update ocean accessible diagonals */
	cartesian_diagonal_iterate(x, y, x1, y1) {
	  if( map_get_terrain(x1,y1) == T_OCEAN ) {
	    ret+= assign_ocean_flood1(x1, y1, n0, nr);
	    nr= ((mm= map_get_ocean(x1, y1)) > 0 ? mm : nr);
	  }
	} cartesian_diagonal_iterate_end;
      }
    }
  }
  return ret;
}

/**************************************************************************
 List the tile and recursively adjacent water tiles with its ocean number.
 It returns the number of tiles added to the list.
 The "list" array has its current count at list[0] and size at list[-1].
**************************************************************************/
static int list_ocean_flood(int *list, int x, int y)
{
  int  m0, n0, nn;

  assert(list[0] < list[-1] );
  n0= list[0];
  m0= map_get_ocean(x, y);
  list_ocean_flood1(list, x, y, m0);

  /* Undo blocking marks ... */
  for(nn= list[0]; nn > n0; nn--) {
    y= list[nn]/map.xsize, x= list[nn]%map.xsize;
    map_set_ocean(x, y, m0);
  }
  return list[0]-n0;
}

/**************************************************************************
 List this tile and recursively adjacent water tiles with number "n0".
 This function flips the sign of the ocean number as each tile is counted.
 The "list" array has its current count at list[0] and size at list[-1].
**************************************************************************/
static int list_ocean_flood1(int *list, int x, int y, int n0)
{
  assert(list[0] < list[-1] );

  if( n0 != 0 && map_get_ocean(x, y) == n0 && list[0] < list[-1] ) {
    /* block this square */
    map_set_ocean(x, y, -n0);
    list[++list[0]]= y*map.xsize+x;
    adjc_iterate(x, y, x1, y1) {
      list_ocean_flood1(list, x1, y1, n0);
    } adjc_iterate_end;
  }
  return list[0];
}

/**************************************************************************
 Assign ocean numbers to all tiles.
 Also sets map_num_oceans() (oceans have numbers 1 to
 map_num_oceans() _inclusive_).
**************************************************************************/
void assign_ocean_numbers(void)
{
  int seas = 0;

  whole_map_iterate(x, y) {
      map_set_ocean(x, y, 0);
  } whole_map_iterate_end;

  whole_map_iterate(x, y) {
    if (!map_get_ocean(x, y) && map_get_terrain(x, y)==T_OCEAN) {
      assign_ocean_flood(x, y, ++seas);
    }
  } whole_map_iterate_end;

  map_num_oceans() = seas;
  freelog(LOG_VERBOSE, "Map has %d oceans", map_num_oceans());
}

/**************************************************************************
 Allocate oceans array and fill in values.
**************************************************************************/
static void setup_seadata(void)
{
  int ii;

  assign_ocean_numbers();
  assert(map_num_oceans()>0);

  /* allocate + 1 so can use ocean number as index */
  if( oceans != NULL )
    free(oceans);
  oceans= fc_malloc((map_num_oceans()+1)*sizeof(struct seadata));

  /* initialize: */
  for(ii= 0; ii <= map_num_oceans(); ii++) {
    oceans[ii].x= oceans[ii].y= -1;	     /* flag */
    oceans[ii].count= oceans[ii].rivers= 0;
  }

  /* get x and y positions: (top left) */
  whole_map_iterate(x, y) {
    int tx= map_get_terrain(x, y);
    ii= map_get_ocean(x, y);
    oceans[ii].count++;
    if (tx == T_RIVER || (map_get_special(x,y)&S_RIVER))
      oceans[ii].rivers++;
    else
    if (oceans[ii].x == -1 && tx == T_OCEAN)
      oceans[ii].x= x, oceans[ii].y= y;
  } whole_map_iterate_end;

  freelog(LOG_VERBOSE, _("Map has %i seas of size: "), map_num_oceans());
  for(ii=1; ii<=map_num_oceans(); ii++) {
    freelog(LOG_VERBOSE, _("%2d %2d,%-2d = %d/%d"),
      ii, oceans[ii].x, oceans[ii].y, oceans[ii].count, oceans[ii].rivers);
  }
}

/**************************************************************************
 Allocate islands array and fill in values.
 Note this is only use for map.generator<=1, since others
 setups islands and starters explicitly.
**************************************************************************/
static void setup_isledata(void)
{
  int good, mingood, maxgood;
  int riches;
  int starters;
  int isles, oldisles, goodisles;
  int guard1=0;
  int firstcont;
  int i;

  assert(map.num_continents>0);

  /* allocate + 1 so can use continent number as index */
  islands = fc_malloc((map.num_continents+1)*sizeof(struct isledata));

  /* initialize: */
  for(i=0; i<=map.num_continents; i++) {
    islands[i].x = islands[i].y = -1;	     /* flag */
    islands[i].goodies = islands[i].starters = 0;
  }

  /* get x and y positions: (top left) */
  whole_map_iterate(x, y) {
    int cont = map_get_continent(x, y);
    if (islands[cont].x == -1) {
      islands[cont].x = x;
      islands[cont].y = y;
    }
  } whole_map_iterate_end;

  /* Add up the goodies: for useable ocean, add value to continent
     for _every_ adjacent land tile.  This is IMO not very good,
     because it adds potentially many times, and usable sea of
     distance 2 from land is ignored, but this re-produces the
     results of the previous method which used flood_fill().  --dwp
     This is also the correct place to add S_HUT bonus.
  */
  whole_map_iterate(x, y) {
    int cont = map_get_continent(x, y);
    if (cont) {
      islands[cont].goodies += is_good_tile(x, y);
      if (map_get_special(x,y) & S_HUT) {
	islands[cont].goodies += 0; /* 3; */   /* regression testing */
      }
    } else {
      assert(map_get_terrain(x, y)==T_OCEAN);
      /* no need to check is_sea_usable(x,y), because will
	 only use for adjacent land (cont1>0) below */
      {
	int goodval = is_good_tile(x, y);
	adjc_iterate(x, y, x1, y1) {
	  int cont1 = map_get_continent(x1, y1);
	  if (cont1>0)
	    islands[cont1].goodies += goodval;
	} adjc_iterate_end;
      }
    }
  } whole_map_iterate_end;

  /* the arctic and the antarctic are continents 1 and 2 for generator>0*/
  if (map.generator>0) {
    firstcont = 3;
  } else {
    firstcont = 1;
  }
  isles = map.num_continents+1;

  riches=0;
  for (i=firstcont; i<isles; i++) {
      riches+=islands[i].goodies;
  }

  starters= 100; oldisles= isles+1; goodisles= isles;
  while( starters>game.nplayers && oldisles>goodisles ){
    freelog(LOG_VERBOSE, "goodisles=%i", goodisles);
    oldisles= goodisles;
    maxgood= 1;
    mingood= riches;
    good=0;
    goodisles=0;
    starters= 0;

    /* goody goody */
    for (i=firstcont;i<isles;i++) {
      islands[i].starters=0;
      if ( islands[i].goodies > (riches+oldisles-1)/oldisles ) {
	good+=islands[i].goodies;
	goodisles++;
	if(mingood>islands[i].goodies)
	  mingood= islands[i].goodies;
	if(maxgood<islands[i].goodies)
	  maxgood= islands[i].goodies;
      }
    }

    if(mingood+1<maxgood/game.nplayers)
      mingood= maxgood/game.nplayers;

    if(goodisles>game.nplayers){
      /* bloody goodies */
      for (i=firstcont;i<isles;i++) {
	if ( (islands[i].goodies*4 > 3*(riches+oldisles-1)/oldisles)
	  &&!(islands[i].goodies > (riches+oldisles-1)/oldisles) ) {
	  good+=islands[i].goodies;
	  goodisles++;
	  if(mingood>islands[i].goodies)
	    mingood= islands[i].goodies;
	}
      }

      /* starters are loosers */
      for (i=firstcont;i<isles;i++) {
	if ( (islands[i].goodies*4 > 3*(riches+oldisles-1)/oldisles)
	  &&!(islands[i].goodies > (riches+oldisles-1)/oldisles) ) {
	  freelog(LOG_VERBOSE, "islands[i].goodies=%i",islands[i].goodies);
	  islands[i].starters= (islands[i].goodies+guard1)/mingood;
	  if(!islands[i].starters) {
	    islands[i].starters+= 1;/* ?PS: may not be enough, guard1(tm) */
	  }
	  starters+= islands[i].starters;
	}
      }
    }

    /* starters are winners */
    for (i=firstcont;i<isles;i++) {
      if (islands[i].goodies > (riches+oldisles-1)/oldisles) {
	assert(!islands[i].starters);
	freelog(LOG_VERBOSE, "islands[i].goodies=%i", islands[i].goodies);
	islands[i].starters = (islands[i].goodies+guard1)/mingood;
	if(!islands[i].starters) {
	  islands[i].starters+= 1;/* ?PS: may not be enough, guard1(tm) */
	}
	starters+= islands[i].starters;
      }
    }

    riches= good;
    if(starters<game.nplayers){
      starters= game.nplayers+1;
      goodisles=oldisles;
      oldisles= goodisles+1;
      riches= (4*riches+3)/3;
      if(mingood/game.nplayers>5) {
	guard1+= mingood/game.nplayers;
      } else {
	guard1+=5;
      }
      freelog(LOG_DEBUG,
	      _("Map generator: not enough start positions, fixing."));
    }
  }
  freelog(LOG_DEBUG, _("The map has %i starting positions on %i isles."),
	  starters, goodisles);
}

/**************************************************************************
  where do the different races start on the map? well this function tries
  to spread them out on the different islands.
**************************************************************************/
#define MAXTRIES 1000000
void create_start_positions(void)
{
  int nr=0;
  int k, sum, dist;
  int tx, x,y, xy,*yx;

  if (islands==NULL)		/* already setup for generator>1 */
    setup_isledata();

  sum=0;
  for (k=0; k<=map.num_continents; k++) {
    sum += islands[k].starters;
    if (islands[k].starters!=0) {
      freelog(LOG_VERBOSE, "starters on isle %i", k);
    }
  }
  assert(game.nplayers<=nr+sum);

  dist=40;     /* sanity */
  if(dist>= map.xsize/3)
    dist= map.xsize/3;
  if(dist>= map.ysize/3)
    dist= map.ysize/3;
  if(dist<=5)
    dist = 5;  /* sanity */

  /* Collect a randomized list of starting tiles */
  xy= map.ysize*map.xsize;
  yx= (int *)fc_malloc(xy*sizeof(int));
  for( k= 0; xy-- > 0; ) {
    y= xy / map.xsize, x= xy % map.xsize;
    if (islands[(int)map_get_continent(x, y)].starters) {
      tx= map_get_terrain(x, y);
      if( tx == T_GRASSLAND
       || tx == T_PLAINS
       || tx == T_HILLS )
	x= myrand(k+1), yx[k++]= yx[x], yx[x]= xy;
    }
  }

  /* Pick off tiles separated by at least dist */
  while(dist>=4 && nr<game.nplayers) {
    for (xy= k; xy-- > 0;) {
      x = yx[xy] % map.xsize;
      y = yx[xy] / map.xsize;
      if (islands[(int)map_get_continent(x, y)].starters) {
	if (!is_starter_close(x, y, nr, dist)) {
	  islands[(int)map_get_continent(x, y)].starters--;
	  map.start_positions[nr].x= x;
	  map.start_positions[nr].y= y;
	  if( ++nr >= game.nplayers )
	    break;
	}
      }
    }
    dist=(dist*8)/10;
  }
  map.num_start_positions = nr;

  free(yx);
  free(islands);
  islands = NULL;
}

/**************************************************************************
 readjust terrain counts so that it makes sense for mapgen 1, 2, 3 and 4
 idea: input is the number of terrain
 mapgen 1 needs custom parameters,
 mapgen 2 and 3 and 4 use percents, currently.
 Ultimately, I hope all parameters below will be weights,
 with the defaults set to a percentage;
 Placing deserts and swamps may cause some problems.

 This function needs to be called from the server everytime a
 a parameter changes.
 It will be called again at game start, too.
**************************************************************************/
void adjust_terrain_param(void)
{
  int total;

  /*!PS: I don't have the time to have several test runs */
  /* to find a mapping from percents to generator 1 settings. */

  if(map.generator==1){
    /*map.riverlength*= 10; */
    /*I leave this out, people will get too upset
      if they have to change all their scripts */
    return;
  }

  map.riverlength/= 10; /* left in */

  /* ignore riverlength
     make_rivers replaces/overrides an amount of existing terrain */
  total = /* map.riverlength + */ map.mountains + map.deserts
    + map.forestsize + map.swampsize + map.grasssize;

  /* recalibrate if things don't add up */
  if(total<90 || total>100) {
    /* map.riverlength= map.riverlength*100/total; */
    map.mountains= map.mountains*100/total;
    map.deserts= map.deserts*100/total;
    map.forestsize= map.forestsize*100/total;
    map.swampsize= map.swampsize*100/total;
  }
  /* residual goes implicitly to grass, so always compute it */
  total = /* map.riverlength + */ map.mountains + map.deserts
    + map.forestsize + map.swampsize + 0;
  map.grasssize= 100 - total;

  map.riverlength*= 10;
}

/**************************************************************************
  This scales heights from mapgenerators > 1 so we can call make_<feature>.
**************************************************************************/
static void adjust_map2()
{
  int nn, tt;

  /* Fixup create_island heights */
  maxval=0;
  whole_map_iterate(x, y) {
    nn= height_map[_xy];
    if( nn<0 ) nn= myrand(100)+3000;    /* polar/blocking paranoia */
    else
    if( nn<100 ) nn= myrand(2000)+1000; /* oceans */
    else {			      /* island landmass */
      nn+= 3000;
      /* terrain assignment does not respect heighs - fake/fix this */
      if((tt= map_get_terrain(x, y))==T_HILLS)
	nn+= myrand(3000)+1000;
      else
      if(tt==T_MOUNTAINS)
	nn+= myrand(6000)+2000;
      if (nn>maxval) maxval=nn;
    }
    height_map[_xy]= nn;

    /* cursory smoothing to stabilize height diffs a little */
    if((nn= _xy+map.xsize+1)<map.ysize*map.xsize && height_map[nn]>=3000) {
      y= nn/map.xsize, x= nn%map.xsize;
      height_map[nn]= (hmap1(x,y,1)+height_map[nn])/2;
    }
  } whole_map_iterate_end

  /* scale 0:tocean(==3000) and tocean:maxval(==7000) */
  /* Note this tends to smooth ocean floors and accentuate land features */
  whole_map_iterate(x,y) {
    if( (nn= height_map[_xy]) >= tocean)
      height_map[_xy]= ((nn-tocean)*7000)/(maxval-tocean)+3000;
  } whole_map_iterate_end

  /* Final stats */
  trough= 0;
  whole_map_iterate(x,y) {
    trough+= (map_get_terrain(x,y)==T_OCEAN)?0:is_rough(x,y);
  } whole_map_iterate_end

  tocean= 3000;
  maxval= 10000;
}

/**************************************************************************
  since the generated map always has a positive number as minimum height
  reduce the height so the lowest height is zero and the maximum is 10000,
  also, set ~sealevel to 3000, this makes calculations easier
**************************************************************************/
static void adjust_map()
{
  int nn, minval;

  tocean= maxval=0;
  minval=32767;
  whole_map_iterate(x, y) {
    tocean+= nn= height_map[_xy];
    if (nn>maxval) maxval=nn;
    else
    if (nn<minval) minval=nn;
  } whole_map_iterate_end;

  /* maxval, average shifted to minval == 0 */
  tocean/=map.ysize*map.xsize;
  tocean=tocean-minval+1;
  maxval=maxval-minval+1;
  if( maxval <= tocean) maxval= tocean+1; /* paranoia */

  /* scale 0:tocean(==3000) and tocean:maxval(==7000) */
  /* Note this tends to smooth ocean floors and accentuate land features */
  whole_map_iterate(x, y) {
    nn= height_map[_xy]-minval;
    if (nn>tocean)
      height_map[_xy]= ((nn-tocean)*7000)/(maxval-tocean)+3000;
    else
      height_map[_xy]= (nn*3000)/tocean;
  } whole_map_iterate_end;

  tocean= 3000;
  maxval= 10000;
}

/**************************************************************************
  smooth_map should be viewed as a corrosion function on the map, it levels
  out the differences in the heightmap.
**************************************************************************/
static void smooth_map(void)
{
  int a0, a1, a2, n1, n2, nn;
  int xy,*yx;

  /* save results in a scratch array to avoid polluting the source */
  xy = map.ysize*map.xsize;
  yx = (int *)fc_malloc(xy*sizeof(int));

  nn= 0;
  whole_map_iterate(x, y) {
    /* city sized and adjacent averages */
    a0= hmap1(x, y, 2);
    a0+= a1= hmap1(x, y, 1);

    /* detect key directions for elevation changes */
    a2= hslope(&n2, x, y, 0)-a1;
    a1= hslope(&n1, x, y, 1)-a1;
    a1= (a1<100)?myrand(200):a1;  /* add noise if too smooth */
    a0+= (a1>-a2)?a1*3:a2;        /* land grows faster than sea erodes */
    a1= hmap(x, y);

    /* build depressions == land lakes */
    if( a1 < hmap(DX1(n2)+x,DY1(n2)+y)
     && a1 < hmap(DX1(n2-1)+x,DY1(n2-1)+y)
     && a1 < hmap(DX1(n2+1)+x,DY1(n2+1)+y) )
      a1= (a1>2000?(a1-2000):0);
    a0+= a1;

    /* track the global adjustments */
    nn-= a1;
    nn+= yx[_xy]= a0/3;
  } whole_map_iterate_end;
  nn/= xy;

  /* update source height_map, adjust for constant mass (i.e. settling) */
  whole_map_iterate(x, y) {
    height_map[_xy]= (yx[_xy]>nn) ? (yx[_xy]-nn):0;
  } whole_map_iterate_end;

  free(yx);
}

/**************************************************************************
  mapgenerator1, highlevel function, that calls all the previous functions
**************************************************************************/
static void mapgenerator1(void)
{
  int x,y,y1, xy, *yx, s0,s1,s2, hh,nn;

  height_map=fc_malloc(map.xsize*map.ysize*sizeof(int));
  ocean_map=fc_malloc(map.xsize*map.ysize*sizeof(int));
  yx=fc_malloc(map.xsize*map.ysize*sizeof(int));

  adjust_terrain_param();

  /* start with base level fill plus polar basins */
  for (y=0;y<map.ysize/2;y++) {
    y1 = map.ysize-1 - y;
    nn = y * 5000 / map.ysize;
    nn = nn < 500 ? nn+1 : 500;
    for (x=0;x<map.xsize;x++) {
      height_map[y*map.xsize+x]=myrand(nn)+nn;
      height_map[y1*map.xsize+x]=myrand(nn)+nn;
    }
  }

  /* sort map coordinates by decreasing height */
  /* first scramble the map coordinates to avoid subtle NS bias */
  xy= map.ysize*map.xsize;
  for( yx[0]= 0, y=1; y<xy; y++)
    x= myrand(y+1), yx[y]= yx[x], yx[x]= y;
  for(y= xy; y-- > 0;) {
    y1= yx[y], hh= height_map[y1];
    for(x= y; ++x < xy && height_map[yx[x]] > hh;)
      yx[x-1]=yx[x];
    yx[x-1]=y1;
  }
  /* pick three "seas" to avoid - but not the deepest points */
  s0= yx[(xy*8)/10 - myrand(xy)/3];
  s1= yx[(xy*8)/10 - myrand(xy)/3];
  s2= yx[(xy*8)/10 - myrand(xy)/3];

  /* asteroid rain on landpercent of the map */
  for (nn=(xy*map.landpercent)/100;nn-->0;) {

    /*  biasing strikes to higher elevations */
    /*  and away from designated major seas  */
    for(x=0, y1=map.ysize; (y1/=2)>0;) {
      x= myrand(xy), y= myrand(xy), x= (x<y)?x:y;
      if( map_dist(yx[x],s0) >= y1
       && map_dist(yx[x],s1) >= y1
       && map_dist(yx[x],s2) >= y1 )
      break;
    }

    /* point hit with directional splash - 20% ejectile */
    hh= myrand(3000)+2000;
    if( myrand(100) < 20 )
      hh= -(hh > height_map[yx[x]] ? height_map[yx[x]]:hh)/2;
    height_map[yx[x]]+= hh;
    y= myrand(8);  /* direction */

    /* half hit in direction y */
    hh/= 2;
    y1=y, y1= DY1(y1)*map.xsize+DX1(y1)+yx[x];
    if(y1>=map.xsize && (xy-1 - y1)>=map.xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);

    /* quarter hit in neighbouring directions */
    hh/= 2;
    y1=y-1, y1= DY1(y1)*map.xsize+DX1(y1)+yx[x];
    if(y1>=map.xsize && (xy-1 - y1)>=map.xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);
    y1=y+1, y1= DY1(y1)*map.xsize+DX1(y1)+yx[x];
    if(y1>=map.xsize && (xy-1 - y1)>=map.xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);

    /* periodically smooth, recalibrate and resort */
    if ((nn%map.xsize) == 0) {
      smooth_map();
      adjust_map();
      for(y=xy; y-->0;) {
	y1= yx[y], hh= height_map[y1];
	for(x=y; ++x<xy && height_map[yx[x]]>hh;)
	  yx[x-1]=yx[x];
	yx[x-1]=y1;
      }
      /* drop one sea and choose a new altantis */
      s0= s1, s1= s2, s2= yx[(xy*8)/10 - myrand(xy)/3];
    }
  }
  smooth_map();
  adjust_map();
  free(yx);

  /* populate the map using the generated height_map data */
  make_land();

  free(ocean_map);
  free(height_map);
  ocean_map= height_map= NULL;
}

/**************************************************************************
  globals for generator 2, 3 & 4
**************************************************************************/
static int  *create_map;
static int  isleindex, n, e, s, w;
static long int totalmass; /* better a global than a duplicate formula */
static long int checkmass;
static long int totalweight;

/**************************************************************************
  fill an island with up four types of terrains, rivers done separately
**************************************************************************/
static void fill_island(long int *bucket, int coast,
	int warm0_weight, int warm1_weight, int cold0_weight, int cold1_weight,
	enum tile_terrain_type warm0, enum tile_terrain_type warm1,
	enum tile_terrain_type cold0, enum tile_terrain_type cold1)
{
  int x, y, k, xx, xy, *yx;

  if( *bucket <= 0
   || (warm0_weight+warm1_weight+cold0_weight+cold1_weight<=0))
    return;

  /* collect a randomized list of eligible locations */
  xx= map.ysize*map.xsize;
  yx= fc_malloc(sizeof(int)*xx);
  for( k= 0, xy= -1; ++xy<xx; ) {
    y= xy/map.xsize, x= xy%map.xsize;
    if (map_get_continent(x,y) == isleindex &&
	map_get_terrain(x,y) == T_GRASSLAND)
      x= myrand(k+1), yx[k++]= yx[x], yx[x]= xy;
  }

  /* loop until quota filled or locations run out */
  xx= *bucket/10 + 2; /* number of (guaranteed) seed locations */
  while (*bucket>0 && k-->0) {
    y= yx[k]/map.xsize, x= yx[k]%map.xsize;

    /* the first condition helps make terrain more contiguous,
       the next four help control feature clumping
       the last lets it avoid the coast. */
    if( ( (is_terrain_near_tile(x,y,warm0)||is_terrain_near_tile(x,y,cold0))
	|| myrand(totalmass)<*bucket
	|| xx-->0 )
     && (!warm0_weight || count_terrain_near_tile(x,y,warm0)<warm0_weight)
     && (!warm1_weight || count_terrain_near_tile(x,y,warm1)<warm1_weight)
     && (!cold0_weight || count_terrain_near_tile(x,y,cold0)<cold0_weight)
     && (!cold1_weight || count_terrain_near_tile(x,y,cold1)<cold1_weight)
     && (!is_at_coast(x, y) || myrand(100) < coast )) {
      if ( is_cold(x,y) )
	(*bucket)--, map_set_terrain(x, y,
	  (myrand(cold0_weight+cold1_weight+2)>=cold1_weight)?cold0:cold1);
      else
	(*bucket)--, map_set_terrain(x, y,
	  (myrand(warm0_weight+warm1_weight+2)>=warm1_weight)?warm0:warm1);
    }
  }
  free(yx);
}

/**************************************************************************
  finds a place and drops the island(s) found in create_map
  forever can be used to limit the number of tries
**************************************************************************/
static int place_island(int *create_map, int forever)
{
  int x, y, xx, yy, xy, *yx, k, xo= 0, yo= 0, i=0;

  /* collect a randomized list of eligible starting squares */
  xx= (map.ysize-(s-n))*map.xsize;
  yx= fc_malloc(sizeof(int)*xx);
  for( k= 0, xy= -1; ++xy<xx; )
    x= myrand(k+1), yx[k++]= yx[x], yx[x]= xy;

  /* loop over the list or requested count, whichever runs out first */
  if(forever==0) forever= k;  /* convenience */
  while(forever-->0 && k-->0) {
    yo= yx[k]/map.xsize, xo= yx[k]%map.xsize;

    /* a diagonal pre-check helps a lot for maps with high landmass */
    for (y= n, x= w; y <= s && x <= e; y++, x++) {
      if (create_map[map.xsize*y+map_adjust_x(x)]) {
	xx= x + xo - w, yy= y + yo - n;
	if (!normalize_map_pos(&xx, &yy) || is_coastline(xx, yy))
	  goto end_while;
      }
    }

    /* full rectangular overlap check */
    block_iterate2(x, y, w, n, e, s) {
      if (create_map[_xy]) {
	xx= x + xo - w, yy= y + yo - n;
	if (!normalize_map_pos(&xx, &yy) || is_coastline(xx, yy))
	  goto end_while;
      }
    } block_iterate2_end;

    /* finally, update the real map(s) */
    block_iterate2(x, y, w, n, e, s) {
      if ((xy= create_map[_xy])) {
	if(checkmass-- <= 0) {
	  freelog(LOG_ERROR, "mapgen.c: mass doesn't sum up.");
	  goto exit_while;
	}
	xx= x + xo - w, yy= y + yo - n;
	map_set_terrain(xx, yy, T_GRASSLAND);
	map_set_continent(xx, yy, isleindex);
	hmap(xx,yy)= xy;
	i++;
      }
    } block_iterate2_end;
    break;
    end_while:
  }
  exit_while:
  free(yx);

  /* reset to real map bounding rectangle */
  s += yo - n;
  e += xo - w;
  n = yo;
  w = xo;

  return i;
}

/**************************************************************************
  create an island centred in a scratch map when called with islemass != 0
**************************************************************************/
static void create_island(int *height_map, int islemass)
{
  int dd[8] = {1000, 500, 300, 200, 500, 200, 300, 500};
  int hh, xx, yy, xy, *yx, j, k, l, m, d0, d1= 1, na= 0, nc=0;

  /* setup list buffer and initialize the scratch height_map */
  xy= map.ysize*map.xsize;
  yx= fc_malloc(sizeof(int) * xy);
  memset(height_map,              -1, sizeof(int)*map.xsize);
  memset(height_map+map.xsize,     0, sizeof(int)*(xy-map.xsize*2));
  memset(height_map+xy-map.xsize, -1, sizeof(int)*map.xsize);

  /* Add a hill at the map centre - bias growth in the d0'th direction */
  xy= map.ysize/2, xy= xy*map.xsize+map.xsize/2;
  yx[na++] = xy;
  height_map[xy]= 7000;
  d0= myrand(8);

  /* Seeding done - now loop */
  while(nc < islemass && na > 0) {
    /* Choose a tile to add - bias choice to higher adjacent */
    j= myrand(na), k= myrand(na), k= (j<k?j:k)+nc;

    /* move it down to the selected set and reset counts */
    xy= yx[k];
    while( k-->nc )
      yx[k+1]= yx[k];
    yx[(na--,nc++)]= xy;
    yy= xy/map.xsize, xx= xy%map.xsize;

    /* Update the directional bias, i.e. non-spheriodal preference  */
    /* in combination with a rapid height dropoff, makes peninsulas */
    /* Note: this uses an octal heartbeat with 1-2-1 change probability */
    if( (nc&7)==0 )
      d1++, d0= ((myrand(4)-1)/2 + d0)&7;

    /* more anti-spherism - start a new coastal growth bump */
    if( d1>1 ) {
      m= 0;
      adjc_iterate(xx, yy, x1, y1) {
	m+= (height_map[map.xsize*y1+x1]==0?1:0);
      } adjc_iterate_end
      /* bump iff really coastal */
      if( m>4 ) {
	d1=0, height_map[xy]= 7000;

	/* ... and deflate current/adjacent fill areas */
	for(k= nc+na; k-->nc;) {
	  m= 0, j= yx[k]/map.xsize, l= yx[k]%map.xsize;
	  adjc_iterate(j, l, x1, y1) {
	    m+= (height_map[map.xsize*y1+x1]==0?((_nn&1)?2:1):0);
	  } adjc_iterate_end
	  /* don't deflate singly connected or less lakes */
	  if( m>2 )
	    height_map[yx[k]]= (height_map[yx[k]]+1)/2;
	}
      }
    }

    /* Add all uncounted adjacent tiles to the sorted selection list */
    hh= height_map[xy], hh= (hh>1000?hh-1000:1);
    adjc_iterate(xx, yy, x1, y1) {
      j= map.xsize*y1+x1;
      if( height_map[j] == 0 && (l= myrand(dd[(_nn-d0)&7])+hh)>200 ) {
	height_map[j]= l;
	for(k= nc+(na++); k-->nc && height_map[yx[k]]<l;)
	  yx[k+1]= yx[k];
	yx[k+1]= j;
      }
    } adjc_iterate_end
  }

  /* Remove leftovers - first the polar blocks */
  xy= map.ysize*map.xsize;
  memset(height_map+xy-map.xsize, 0, sizeof(int)*map.xsize);
  memset(height_map,              0, sizeof(int)*map.xsize);

  /* ... and adjacent but uncounted */
  for(k=nc+na; k-->nc;)
    height_map[yx[k]]= 0;

  /* Set nwse extents, i.e. bounding rectangle */
  /* TODO: get rid of n, w, s, e single char globals, use a struct arg */
  n= map.ysize, w=map.xsize,  s= e= 0;
  for(l=0, k++; k-->0;) {
    l+= (height_map[yx[k]]?1:0);

    xy= yx[k]/map.xsize;
    if( xy < n ) n= xy;
    else
    if( xy > s ) s= xy;

    xx= yx[k]%map.xsize;
    if( xx < w ) w= xx;
    else
    if( xx > e ) e= xx;
  }
  free(yx);

  if(nc<islemass) {
    /*  (-: NEVER REACHED :-)  */
    freelog(LOG_ERROR, "create_island ended early with %d/%d.",
	    nc, islemass);
  }
}

/**************************************************************************
  make an island, fill every tile type except plains
  note: you have to create big islands first.
**************************************************************************/
static void make_island(int islemass, int starters)
{
  static long int tilefactor, balance, lastplaced; /* int may be only 2 byte! */
  static long int riverbuck, mountbuck, desertbuck, forestbuck, swampbuck;
  int i= 0;

  if (!islemass) {
    balance= 0;     /* setup initial static parameters */
    isleindex= 3;   /* 0= none, 1= arctic, 2= antarctic */
    checkmass= totalmass;

    /* caveat: this should really be sent to all players */
    if(totalmass>3000)
      freelog(LOG_NORMAL, _("High landmass - this may take a few seconds."));

    /* ignore rivers - make_rivers overrides current terrain
    i= (terrain_control.river_style==R_AS_TERRAIN)?(map.riverlength/10):0;
    */
    i+= map.mountains + map.deserts + map.forestsize + map.swampsize
      + map.grasssize;
    tilefactor = i-5;  /* overbudget 5% - smaller islands get dregs */
    riverbuck  = 0;
    mountbuck  = 0;
    desertbuck = 0;
    forestbuck = 0;
    swampbuck  = 0;
    lastplaced = totalmass;
  } else {

    /* makes the islands here */
    if (islemass <= balance) return;

    /* don't create continents without a number */
    if(isleindex >= MAP_NCONT)
      return;
    islands[isleindex].starters = starters;
    freelog(LOG_VERBOSE, "island %i",isleindex);

    /*don't create big isles we can't place*/
    islemass = islemass - balance;
    if(islemass>lastplaced+1+lastplaced/50)
      islemass= lastplaced+1+lastplaced/50;

    i = islemass;
    do {
      create_island(create_map, i--);
      lastplaced= place_island(create_map, 0);
    } while (!lastplaced && i*10>islemass );

    if(lastplaced)
      balance = lastplaced - islemass;
    else
      balance = 0, lastplaced = i;

    freelog(LOG_VERBOSE, "ini=%d, plc=%d, bal=%ld, tot=%ld",
	    islemass, i, balance, checkmass);

    mountbuck += (map.mountains*i)/tilefactor;
    fill_island(&mountbuck, 30, 5,3, 5,3,
	  T_HILLS, T_MOUNTAINS, T_HILLS, T_MOUNTAINS);

    desertbuck += (map.deserts*lastplaced)/tilefactor;
    fill_island(&desertbuck, 20, 6,0, 7,3,
	  T_DESERT, T_DESERT, T_TUNDRA, T_ARCTIC);

    forestbuck += (map.forestsize*lastplaced)/tilefactor;
    fill_island(&forestbuck, 60, 7,4, 4,2,
	  T_FOREST, T_JUNGLE, T_FOREST, T_SWAMP);

    swampbuck += (map.swampsize*lastplaced)/tilefactor;
    fill_island(&swampbuck, 80, 6,2, 6,2,
	  T_SWAMP, T_JUNGLE, T_SWAMP, T_HILLS);

    isleindex++;
    map.num_continents++;
  }
}

/**************************************************************************
  allocate working arrays (caller must free), fill ocean and make polar
**************************************************************************/
static void initworld(void)
{
  int x;

  height_map = fc_malloc(sizeof(int) * map.ysize * map.xsize);
  create_map = fc_malloc(sizeof(int) * map.ysize * map.xsize);
  ocean_map = fc_malloc(sizeof(int) * map.ysize * map.xsize);
  islands = fc_malloc((MAP_NCONT+1)*sizeof(struct isledata));

  /* ocean */
  whole_map_iterate(x, y) {
    map_set_terrain(x, y, T_OCEAN);
    map_set_continent(x, y, 0);
    hmap(x, y)= 0;
  } whole_map_iterate_end;

  /* polar */
  for (x= 0 ; x < map.xsize; x++) {
    map_set_terrain(x, 0, myrand(9) ? T_ARCTIC : T_TUNDRA);
    map_set_continent(x, 0, 1);
    if (!myrand(9)) {
      map_set_terrain(x, 1, myrand(9) ? T_TUNDRA : T_ARCTIC);
      map_set_continent(x, 1, 1);
    }
    hmap(x,0)= hmap(x,1)= tocean+100;
    map_set_terrain(x, map.ysize-1, myrand(9) ? T_ARCTIC : T_TUNDRA);
    map_set_continent(x, map.ysize-1, 2);
    if (!myrand(9)) {
      map_set_terrain(x, map.ysize-2, myrand(9) ? T_TUNDRA : T_ARCTIC);
      map_set_continent(x, map.ysize-2, 2);
    }
    hmap(x,map.ysize-1)= hmap(x,map.ysize-2)= tocean+100;
  }
  map.num_continents = 2;

  /* basic housekeeping */
  make_island(0, 0);
  islands[2].starters = 0;
  islands[1].starters = 0;
  islands[0].starters = 0;
}

/**************************************************************************
  island base map generators
**************************************************************************/
static void mapgenerator2(void)
{
  int i;
  int spares= 1;
  /* constant to make up for islands actually needing additional space
     i.e. ocean boundaries and internal seas */

  if (map.landpercent > 85) {
    map.generator = 1;
    return;
  }

  adjust_terrain_param();
  totalmass= ((map.ysize-6-spares)*(map.xsize-spares) * map.landpercent)/100;

  /*!PS: The weights NEED to sum up to totalweight (dammit) */
  /* copying the flow of the make_island loops is the safest way */
  totalweight= 0;
  for (i = game.nplayers ; i ; i--)
    totalweight+= 70;
  for (i = game.nplayers ; i ; i--)
    totalweight+= 20;
  for (i = game.nplayers ; i ; i--)
    totalweight+= 10;

  initworld();

  for (i = game.nplayers ; i ; i--)
    make_island(70*totalmass/totalweight, 1);
  for (i = game.nplayers ; i ; i--)
    make_island(20*totalmass/totalweight, 0);
  for (i = game.nplayers ; i ; i--)
    make_island(10*totalmass/totalweight, 0);

  /* final cleanups using mapgenerator1 routines */
  adjust_map2();
  make_fair();
  make_plains();
  make_rivers();
  remove_tiny_islands();
  remove_tiny_lakes();
  assign_continent_numbers();
  setup_seadata();
  map_check();

  /* release initworld allocations */
  free(ocean_map);
  free(create_map);
  free(height_map);
  ocean_map= create_map= height_map= NULL;

  if(checkmass>map.xsize+map.ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }
}

/**************************************************************************
On popular demand, this tries to mimic the generator 3 as best as possible.
**************************************************************************/
static void mapgenerator3(void)
{
  int spares= 1;
  int j=0;

  long int islandmass,landmass, size;
  long int maxmassdiv6=20;
  int bigislands;

  if( map.xsize < 40 || map.ysize < 40 || map.landpercent>80 ) {
    freelog(LOG_NORMAL,_("Falling back to generator 2."));
    map.generator = 2;
    return;
  }

  adjust_terrain_param();
  totalmass= ((map.ysize-6-spares)*(map.xsize-spares)*map.landpercent)/100;

  /* subtracting the arctics */
  landmass= ((map.ysize-6)*map.xsize*map.landpercent)/100;
  if( landmass>3*map.ysize+game.nplayers*3 ){
    landmass-= 3*map.ysize;
  }
  bigislands= game.nplayers;

  islandmass= (landmass)/(3*bigislands);
  if(islandmass<4*maxmassdiv6 )
    islandmass= (landmass)/(2*bigislands);
  if(islandmass<3*maxmassdiv6 && game.nplayers*2<landmass )
    islandmass= (landmass)/(bigislands);

  if(islandmass<2)
    islandmass= 2;
  if(islandmass>maxmassdiv6*6)
    islandmass= maxmassdiv6*6;/* !PS: let's try this */

  initworld();

  while( isleindex-2<=bigislands && checkmass>islandmass && ++j<500 ) {
    make_island(islandmass,1);
  }

  if(j==500){
    freelog(LOG_NORMAL, _("Generator 3 didn't place all big islands."));
  }

  islandmass= (islandmass*11)/8;
  /*!PS: I'd like to mult by 3/2, but starters might make trouble then*/
  if(islandmass<2)
    islandmass= 2;


  while( isleindex<=MAP_NCONT-20 && checkmass>islandmass && ++j< 1500 ) {
      if(j<1000)
	size = myrand((islandmass+1)/2+1)+islandmass/2;
      else
	size = myrand((islandmass+1)/2+1);
      if(size<2) size=2;

      make_island(size, (isleindex-2<=game.nplayers)?1:0 );
  }

  /* final cleanups using mapgenerator1 routines */
  adjust_map2();
  make_fair();
  make_plains();
  make_rivers();
  remove_tiny_islands();
  remove_tiny_lakes();
  assign_continent_numbers();
  setup_seadata();
  map_check();

  /* release initworld allocations */
  free(ocean_map);
  free(create_map);
  free(height_map);
  ocean_map= create_map= height_map= NULL;

  if(j==1500) {
    freelog(LOG_NORMAL,_("Generator 3 left %li landmass unplaced."),checkmass);
  } else if(checkmass>map.xsize+map.ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }

}

/**************************************************************************
...
**************************************************************************/
static void mapgenerator4(void)
{
  int bigweight=70;
  int spares= 1;
  int i;

  /* no islands with mass >> sqr(min(xsize,ysize)) */

  i = game.nplayers / 2;
  if ( game.nplayers<2 || map.landpercent > 80) {
    map.generator = 3;
    return;
  }

  if(map.landpercent>60)
    bigweight=30;
  else if(map.landpercent>40)
    bigweight=50;
  else
    bigweight=70;

  spares= (map.landpercent-5)/30;

  adjust_terrain_param();
  totalmass=((map.ysize-6-spares)*(map.xsize-spares)*map.landpercent)/100;

  /*!PS: The weights NEED to sum up to totalweight (dammit) */
  totalweight= 0;
  if (game.nplayers % 2)
    totalweight+= bigweight*3;
  else
    i++;
  while (--i)
    totalweight+= bigweight*2;
  for (i = game.nplayers ; i ; i--)
    totalweight+= 20;
  for (i = game.nplayers ; i ; i--)
    totalweight+= 10;

  initworld();

  i = game.nplayers / 2;
  if (game.nplayers % 2)
    make_island(bigweight*3, 3);
  else
    i++;
  while (--i)
    make_island(bigweight*2*totalmass/totalweight, 2);
  for (i = game.nplayers ; i ; i--)
    make_island(20*totalmass/totalweight, 0);
  for (i = game.nplayers ; i ; i--)
    make_island(10*totalmass/totalweight, 0);

  /* final cleanups using mapgenerator1 routines */
  adjust_map2();
  make_fair();
  make_plains();
  make_rivers();
  remove_tiny_islands();
  remove_tiny_lakes();
  assign_continent_numbers();
  setup_seadata();
  map_check();

  /* release initworld allocations */
  free(ocean_map);
  free(create_map);
  free(height_map);
  ocean_map= create_map= height_map= NULL;

  if(checkmass>map.xsize+map.ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }
}

/**************************************************************************
  See stdinhand.c for information on map generation methods.

FIXME: Some continent numbers are unused at the end of this function, fx
       removed completely by remove_tiny_islands.
       When this function is finished various data is written to "islands",
       indexed by continent numbers, so a simple renumbering would not
       work...
**************************************************************************/
void map_fractal_generate(void)
{
  /* save the current random state: */
  RANDOM_STATE rstate = get_myrand_state();

  if (map.seed==0)
    map.seed = (myrand(MAX_UINT32) ^ time(NULL)) & (MAX_UINT32 >> 1);

  mysrand(map.seed);

  /* don't generate tiles with mapgen==0 as we've loaded them from file */
  /* also, don't delete (the handcrafted!) tiny islands in a scenario */
  if (map.generator != 0) {
    map_allocate();
    /* if one mapgenerator fails, it will choose another mapgenerator */
    /* with a lower number to try again */
    if (map.generator == 4 )
      mapgenerator4();
    if (map.generator == 3 )
      mapgenerator3();
    if( map.generator == 2 )
      mapgenerator2();
    if( map.generator == 1 )
      mapgenerator1();
  }

  if(!map.have_specials) /* some scenarios already provide specials */
    add_specials(map.riches); /* hvor mange promiller specials oensker vi*/

  if (!map.have_huts)
    make_huts(map.huts); /* Vi vil have store promiller; man kan aldrig faa
			    for meget oel! */

  /* restore previous random state: */
  set_myrand_state(rstate);
}

static void map_check(void)
{
  int tx;
  whole_map_iterate(x, y) {
    int cont = map_get_continent(x, y);
    int sea  = map_get_ocean(x, y);
    if ((tx= map_get_terrain(x, y)) == T_OCEAN) {
      if(cont != 0)
	freelog(LOG_ERROR, "Error: x,y,tx,cont=(%d,%d,%d,%d)\n", x,y,tx,cont);
      if( sea == 0)
	freelog(LOG_ERROR, "Error: x,y,tx,sea=(%d,%d,%d,%d)\n", x,y,tx,sea);
      adjc_iterate(x, y, x1, y1) {
	if ((cont= map_get_ocean(x1, y1)) != 0 && cont != sea
	  && map_get_terrain(x1, y1) != T_RIVER
	  && !(map_get_special(x1, y1) & S_RIVER))
	  freelog(LOG_ERROR,
            "Error: x,y,sea=(%d,%d,%d) x1,y1,sea1=(%d,%d,%d)\n",
	    x,y,sea, x1,y1,cont);
      } adjc_iterate_end;
    } else {
      if(cont == 0)
	freelog(LOG_ERROR, "Error: x,y,tx,cont=(%d,%d,%d,%d)\n", x,y,tx,cont);
      if( tx != T_RIVER && !(map_get_special(x,y)&S_RIVER) && sea!=0)
	freelog(LOG_ERROR, "Error: x,y,tx,sea=(%d,%d,%d,%d)\n", x,y,tx,sea);
      adjc_iterate(x, y, x1, y1) {
	if (map_get_terrain(x1, y1) != T_OCEAN
	 && (sea= map_get_continent(x1, y1)) != cont)
	  freelog(LOG_ERROR,
            "Error: x,y,cont=(%d,%d,%d) x1,y1,cont1=(%d,%d,%d)\n",
	    x,y,cont, x1,y1,sea);
      } adjc_iterate_end;
    }
  } whole_map_iterate_end;
}

void map_printstats()
{
  int  ix[] = {0,11,1,10,5,2,6,4,8,3,9,7,12,13,14,15};
  char tx[] = "adfghjm~prstXDEF";
  char buf[128];
  int  xx[16];
  int  nn, tt;

  freelog(LOG_NORMAL,
    "Stats: maxval=%d thill=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thill, tocean, trough, map.num_continents, map_num_oceans());

  for (tt=16;tt-->0;) xx[tt]=0;
  whole_map_iterate(x,y) {
    xx[9]+= ((map_get_special(x, y)&S_RIVER)?1:0);
    xx[(tt= map_get_terrain(x,y))&15]++;
    if( tt > 15 )
      freelog(LOG_ERROR, "*** Bad terrain type (%d,%d)=%4.4x", x, y, tt);
  } whole_map_iterate_end

  buf[0]=0;
  for (tt=12;tt-->0;) {
    nn= strlen(buf);
    my_snprintf(buf+nn, 128-nn, "%c=%4d, ", tx[ix[tt]], xx[ix[tt]]);
  }
  nn= strlen(buf), buf[nn-2]=0;
  freelog(LOG_NORMAL, buf);
}

/* Debugging routines */

static void print_hmap(int *height_map)
{
  char tx[] = ".abcdefghijklmnopqrstuvwxyz";
  int  x,y, nn;

  printf("\nHmap: maxval=%d thill=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thill, tocean, trough, map.num_continents, map_num_oceans());
  for (y=0; y<map.ysize; y++) {
    for (x=0; x<map.xsize; x++)
      if((nn= hmap(x,y))<0)
	putchar(tx[25]);
      else
      if(nn<tocean)
	putchar(tx[((tocean-nn)*25)/(tocean+1)]);
      else
      if(nn<maxval)
	putchar(tx[((nn-tocean)*25)/(maxval-tocean)+1]-32);
      else
	putchar(tx[25]-32);
    putchar('\n');
  }
}

static void print_tmap()
{
  char tx[] = "adfghjm.prstX000";
  int  x,y;

  printf("\nTmap: maxval=%d thill=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thill, tocean, trough, map.num_continents, map_num_oceans());
  for (y=0; y<map.ysize; y++) {
    for (x=0; x<map.xsize; x++)
      putchar(tx[map_get_terrain(x,y)&15]);
    putchar('\n');
  }
}

static void print_fmap()
{
  char tx[] = "-123456789abcdefghijklmnopqrstuvwxyz";
  int  x,y;

  printf("\nFmap: maxval=%d thill=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thill, tocean, trough, map.num_continents, map_num_oceans());
  for (y=0; y<map.ysize; y++) {
    for (x=0; x<map.xsize; x++)
      putchar(tx[map_get_ocean(x,y)]);
    putchar('\n');
  }
}

